"""Tests for the basic commit convention."""

from __future__ import annotations

import pytest

from git_changelog import Commit, LinuxConvention


def test_linux_convention_breaking_change() -> None:
    """Breaking change (singular) is correctly identified."""
    subject = "foo: Added a new breaking feature"
    body = ["BREAKING CHANGE: there is a breaking feature in this code"]
    commit = Commit(
        commit_hash="aaaaaaa",
        subject=subject,
        body=body,
        author_date="1574340645",
        committer_date="1574340645",
    )
    convention = LinuxConvention()
    commit_dict = convention.parse_commit(commit)
    assert commit_dict["is_major"]
    assert not commit_dict["is_minor"]
    assert not commit_dict["is_patch"]


def test_linux_convention_breaking_changes() -> None:
    """Breaking changes (plural) are correctly identified."""
    subject = "bar: Added a new breaking feature"
    body = ["BREAKING CHANGES: there is a breaking feature in this code"]
    commit = Commit(
        commit_hash="aaaaaaa",
        subject=subject,
        body=body,
        author_date="1574340645",
        committer_date="1574340645",
    )
    convention = LinuxConvention()
    commit_dict = convention.parse_commit(commit)
    assert commit_dict["is_major"]
    assert not commit_dict["is_minor"]
    assert not commit_dict["is_patch"]


@pytest.mark.parametrize(
    ("subject", "expected_is_major", "expected_is_minor", "expected_is_patch", "expected_type"),
    [
        # Test all types
        ("baz: Added a new feature", False, True, False, "Added"),
        ("baz: Add a new feature", False, True, False, "Added"),
        ("baz: Fix a problem", False, False, True, "Fixed"),
        ("baz: Change implementation", False, False, True, "Changed"),
        ("baz: Remove config file", False, False, True, "Removed"),
        ("baz: Merge branch abc", False, False, True, "Merged"),
        ("baz: Document a new feature", False, False, True, "Documented"),
        ("foo: Implement new API endpoint", False, False, True, ""),
        # Missing space
        ("bar:Add support for new protocol", False, True, False, "Added"),
        # Without "scope:"
        ("Add support for new protocol", False, True, False, "Added"),
        ("Fix typo in documentation", False, False, True, "Fixed"),
        ("Implement new API endpoint", False, False, True, ""),
    ],
)
def test_linux_convention(
    subject: str,
    expected_is_major: bool,
    expected_is_minor: bool,
    expected_is_patch: bool,
    expected_type: str,
) -> None:
    """Commit messages are correctly identified."""
    commit = Commit(
        commit_hash="aaaaaaa",
        subject=subject,
        author_date="1574340645",
        committer_date="1574340645",
    )
    convention = LinuxConvention()
    commit_dict = convention.parse_commit(commit)
    assert commit_dict["is_major"] == expected_is_major
    assert commit_dict["is_minor"] == expected_is_minor
    assert commit_dict["is_patch"] == expected_is_patch
    assert commit_dict["type"] == expected_type



"""Tests for the conventional commit convention."""

from __future__ import annotations

from git_changelog import Commit, ConventionalCommitConvention


def test_conventional_convention_breaking_change() -> None:
    """Breaking change (singular) is correctly identified."""
    subject = "feat: this is a new breaking feature"
    body = ["BREAKING CHANGE: there is a breaking feature in this code"]
    commit = Commit(
        commit_hash="aaaaaaa",
        subject=subject,
        body=body,
        author_date="1574340645",
        committer_date="1574340645",
    )
    convention = ConventionalCommitConvention()
    commit_dict = convention.parse_commit(commit)
    assert commit_dict["type"] == "Features"
    assert commit_dict["scope"] is None
    assert commit_dict["is_major"]
    assert not commit_dict["is_minor"]
    assert not commit_dict["is_patch"]


def test_conventional_convention_breaking_changes() -> None:
    """Breaking changes (plural) are correctly identified."""
    subject = "feat: this is a new breaking feature"
    body = ["BREAKING CHANGES: there is a breaking feature in this code"]
    commit = Commit(
        commit_hash="aaaaaaa",
        subject=subject,
        body=body,
        author_date="1574340645",
        committer_date="1574340645",
    )
    convention = ConventionalCommitConvention()
    commit_dict = convention.parse_commit(commit)
    assert commit_dict["type"] == "Features"
    assert commit_dict["scope"] is None
    assert commit_dict["is_major"]
    assert not commit_dict["is_minor"]
    assert not commit_dict["is_patch"]


def test_conventional_convention_subject_breaking_change() -> None:
    """Breaking change in the subject (!) are correctly identified."""
    subject = "feat!: this is a new breaking feature"
    body = ["There is a breaking feature in this code"]
    commit = Commit(
        commit_hash="aaaaaaa",
        subject=subject,
        body=body,
        author_date="1574340645",
        committer_date="1574340645",
    )
    convention = ConventionalCommitConvention()
    commit_dict = convention.parse_commit(commit)
    assert commit_dict["type"] == "Features"
    assert commit_dict["scope"] is None
    assert commit_dict["is_major"]
    assert not commit_dict["is_minor"]
    assert not commit_dict["is_patch"]


def test_conventional_convention_subject_breaking_change_with_scope() -> None:
    """Breaking change in the subject (!) with scope are correctly identified."""
    subject = "feat(scope)!: this is a new breaking feature"
    body = ["There is a breaking feature in this code"]
    commit = Commit(
        commit_hash="aaaaaaa",
        subject=subject,
        body=body,
        author_date="1574340645",
        committer_date="1574340645",
    )
    convention = ConventionalCommitConvention()
    commit_dict = convention.parse_commit(commit)
    assert commit_dict["type"] == "Features"
    assert commit_dict["scope"] == "scope"
    assert commit_dict["is_major"]
    assert not commit_dict["is_minor"]
    assert not commit_dict["is_patch"]


def test_conventional_convention_feat() -> None:
    """Feature commit is correctly identified."""
    subject = "feat: this is a new feature"
    commit = Commit(
        commit_hash="aaaaaaa",
        subject=subject,
        author_date="1574340645",
        committer_date="1574340645",
    )
    convention = ConventionalCommitConvention()
    commit_dict = convention.parse_commit(commit)
    assert commit_dict["type"] == "Features"
    assert commit_dict["scope"] is None
    assert not commit_dict["is_major"]
    assert commit_dict["is_minor"]
    assert not commit_dict["is_patch"]


def test_conventional_convention_feat_with_scope() -> None:
    """Feature commit is correctly identified."""
    subject = "feat(scope): this is a new feature"
    commit = Commit(
        commit_hash="aaaaaaa",
        subject=subject,
        author_date="1574340645",
        committer_date="1574340645",
    )
    convention = ConventionalCommitConvention()
    commit_dict = convention.parse_commit(commit)
    assert commit_dict["type"] == "Features"
    assert commit_dict["scope"] == "scope"
    assert not commit_dict["is_major"]
    assert commit_dict["is_minor"]
    assert not commit_dict["is_patch"]


def test_conventional_convention_fix() -> None:
    """Bug fix commit is correctly identified."""
    subject = "fix: this is a bug fix"
    commit = Commit(
        commit_hash="aaaaaaa",
        subject=subject,
        author_date="1574340645",
        committer_date="1574340645",
    )
    convention = ConventionalCommitConvention()
    commit_dict = convention.parse_commit(commit)
    assert commit_dict["type"] == "Bug Fixes"
    assert commit_dict["scope"] is None
    assert not commit_dict["is_major"]
    assert not commit_dict["is_minor"]
    assert commit_dict["is_patch"]


def test_conventional_convention_fix_with_scope() -> None:
    """Bug fix commit is correctly identified."""
    subject = "fix(scope): this is a bug fix"
    commit = Commit(
        commit_hash="aaaaaaa",
        subject=subject,
        author_date="1574340645",
        committer_date="1574340645",
    )
    convention = ConventionalCommitConvention()
    commit_dict = convention.parse_commit(commit)
    assert commit_dict["type"] == "Bug Fixes"
    assert commit_dict["scope"] == "scope"
    assert not commit_dict["is_major"]
    assert not commit_dict["is_minor"]
    assert commit_dict["is_patch"]



"""Configuration for the pytest test suite."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from git_changelog._internal.commit import AngularConvention
from tests.helpers import GitRepo

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_gitconfig import GitConfig


@pytest.fixture(name="gitconfig", scope="session")
def _default_gitconfig(default_gitconfig: GitConfig) -> GitConfig:
    default_gitconfig.set({"user.name": "dummy"})
    default_gitconfig.set({"user.email": "dummy@example.com"})
    return default_gitconfig


@pytest.fixture(name="repo")
def git_repo(tmp_path: Path, request: pytest.FixtureRequest, gitconfig: GitConfig) -> GitRepo:  # noqa: ARG001
    """Pytest fixture setting up a temporary Git repository.

    Parameters:
        tmp_path: Path to a temporary directory (pytest fixture).
        request: Pytest fixture request object.
        gitconfig: Git configuration fixture.

    Yields:
        A Git repository wrapper instance.
    """
    repo = GitRepo(tmp_path)
    versions = getattr(request, "param", [])
    for version in versions:
        for section in AngularConvention.TYPES:
            repo.commit(f"{section}: Summary.")
        if version:
            repo.git("tag", version)
    return repo



"""Deprecated. Import from `git_changelog` directly."""

# YORE: Bump 3: Remove file.

import warnings
from typing import Any

from git_changelog._internal import cli


def __getattr__(name: str) -> Any:
    warnings.warn(
        "Importing from `git_changelog.cli` is deprecated. Import from `git_changelog` directly.",
        DeprecationWarning,
        stacklevel=2,
    )
    return getattr(cli, name)



# Module containing the commit logic.

from __future__ import annotations

import re
import sys
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import suppress
from datetime import datetime, timezone
from re import Pattern
from typing import TYPE_CHECKING, Any, ClassVar, SupportsIndex, overload

if sys.version_info < (3, 13):
    from typing_extensions import deprecated
else:
    from warnings import deprecated

if TYPE_CHECKING:
    from collections.abc import Callable, ItemsView, Iterable, KeysView, ValuesView

    from git_changelog._internal.providers import ProviderRefParser, Ref
    from git_changelog._internal.versioning import ParsedVersion


def _clean_body(lines: list[str]) -> list[str]:
    while lines and not lines[0].strip():
        lines.pop(0)
    while lines and not lines[-1].strip():
        lines.pop()
    return lines


def _is_valid_version(version: str, version_parser: Callable[[str], tuple[ParsedVersion, str]]) -> bool:
    try:
        version_parser(version)
    except ValueError:
        return False
    return True


# YORE: Bump 3: Remove block.
class _Trailers(list[tuple[str, str]]):
    def __init__(self, values: Iterable[tuple[str, str]] | None = None):
        super().__init__(values or ())

    @property
    def _dict(self) -> dict[str, str]:
        return dict(iter(self))

    def __contains__(self, key: str | tuple[str, str]) -> bool:  # ty:ignore[invalid-method-override]
        if isinstance(key, str):
            warnings.warn(
                "Checking membership of a string in trailers will stop being supported in version 3.",
                DeprecationWarning,
                stacklevel=2,
            )
            return key in (title for title, _ in self)
        return super().__contains__(key)

    @deprecated("Trailers are now a list of 2-tuples.", category=DeprecationWarning, stacklevel=2)
    def items(self) -> ItemsView:
        return self._dict.items()

    @deprecated("Trailers are now a list of 2-tuples.", category=DeprecationWarning, stacklevel=2)
    def keys(self) -> KeysView:
        return self._dict.keys()

    @deprecated("Trailers are now a list of 2-tuples.", category=DeprecationWarning, stacklevel=2)
    def values(self) -> ValuesView:
        return self._dict.values()

    @deprecated("Trailers are now a list of 2-tuples.", category=DeprecationWarning, stacklevel=2)
    def get(self, key: str, default: str | None = None) -> str | None:
        return self._dict.get(key, default)

    @overload
    def __getitem__(self, key: str) -> str: ...

    @overload
    def __getitem__(self, key: SupportsIndex) -> tuple[str, str]: ...

    @overload
    def __getitem__(self, key: slice) -> list[tuple[str, str]]: ...

    def __getitem__(self, key: str | SupportsIndex | slice) -> str | tuple[str, str] | list[tuple[str, str]]:
        if isinstance(key, str):
            warnings.warn(
                "Getting a trailer with a string key will stop being supported in version 3. Instead, use an integer index, or iterate.",
                DeprecationWarning,
                stacklevel=2,
            )
            return self._dict[key]
        return super().__getitem__(key)

    @overload
    def __setitem__(self, key: str, value: str) -> None: ...

    @overload
    def __setitem__(self, key: SupportsIndex, value: tuple[str, str]) -> None: ...

    @overload
    def __setitem__(self, key: slice, value: Iterable[tuple[str, str]]) -> None: ...

    def __setitem__(
        self,
        key: str | SupportsIndex | slice,
        value: str | tuple[str, str] | Iterable[tuple[str, str]],
    ) -> None:
        if isinstance(key, str):
            warnings.warn(
                "Setting a trailer with a string key will stop being supported in version 3. Instead, append a 2-tuple.",
                DeprecationWarning,
                stacklevel=2,
            )
            self.append((key, value))  # ty:ignore[invalid-argument-type]
            return
        super().__setitem__(key, value)  # ty:ignore[no-matching-overload]


class Commit:
    """A class to represent a commit."""

    def __init__(
        self,
        commit_hash: str,
        author_name: str = "",
        author_email: str = "",
        author_date: str | datetime = "",
        committer_name: str = "",
        committer_email: str = "",
        committer_date: str | datetime = "",
        refs: str = "",
        subject: str = "",
        body: list[str] | None = None,
        url: str = "",
        *,
        parse_trailers: bool = False,
        parent_hashes: str | list[str] = "",
        commits_map: dict[str, Commit] | None = None,
        version_parser: Callable[[str], tuple[ParsedVersion, str]] | None = None,
    ):
        """Initialization method.

        Parameters:
            commit_hash: The commit hash.
            author_name: The author name.
            author_email: The author email.
            author_date: The authoring date (datetime, UTC timestamp with optional timezone offset).
            committer_name: The committer name.
            committer_email: The committer email.
            committer_date: The committing date (datetime, UTC timestamp with optional timezone offset).
            refs: The commit refs.
            subject: The commit message subject.
            body: The commit message body.
            url: The commit URL.
            parse_trailers: Whether to parse Git trailers.
        """
        if not author_date:
            author_date = datetime.now().astimezone()
        elif isinstance(author_date, str):
            epoch, _, offset = author_date.partition(" ")
            tz = datetime.strptime(offset, "%z").tzinfo if offset else timezone.utc
            author_date = datetime.fromtimestamp(float(epoch), tz=tz)
        if not committer_date:
            committer_date = datetime.now().astimezone()
        elif isinstance(committer_date, str):
            epoch, _, offset = committer_date.partition(" ")
            tz = datetime.strptime(offset, "%z").tzinfo if offset else timezone.utc
            committer_date = datetime.fromtimestamp(float(epoch), tz=tz)

        self.hash: str = commit_hash
        """The commit hash."""
        self.author_name: str = author_name
        """The author name."""
        self.author_email: str = author_email
        """The author email."""
        self.author_date: datetime = author_date
        """The author date."""
        self.committer_name: str = committer_name
        """The committer name."""
        self.committer_email: str = committer_email
        """The committer email."""
        self.committer_date: datetime = committer_date
        """The committer date."""
        self.subject: str = subject
        """The commit subject."""
        self.body: list[str] = _clean_body(body) if body else []
        """The commit body."""
        self.url: str = url
        """The commit URL."""

        tag = ""
        for ref in refs.split(","):
            ref = ref.strip()  # noqa: PLW2901
            if ref.startswith("tag: "):
                ref = ref.replace("tag: ", "")  # noqa: PLW2901
                if version_parser is None or _is_valid_version(ref, version_parser):
                    tag = ref
                    break
        self.tag: str = tag
        """The commit tag."""
        self.version: str = tag
        """The commit version."""

        if isinstance(parent_hashes, str):
            parent_hashes = parent_hashes.split()
        self.parent_hashes = parent_hashes
        """The parent commit hashes."""
        self._commits_map = commits_map

        self.text_refs: dict[str, list[Ref]] = {}
        """The commit text references."""
        self.convention: dict[str, Any] = {}
        """The commit message convention."""

        # YORE: Bump 3: Replace `_Trailers()` with `[]` within line.
        self.trailers: list[tuple[str, str]] = _Trailers()
        """The commit trailers."""
        self.body_without_trailers = self.body
        """The commit body without trailers."""

        if parse_trailers:
            self._parse_trailers()

    @property
    def parent_commits(self) -> list[Commit]:
        """Parent commits of this commit."""
        if not self._commits_map:
            return []
        return [
            self._commits_map[parent_hash] for parent_hash in self.parent_hashes if parent_hash in self._commits_map
        ]

    def update_with_convention(self, convention: CommitConvention) -> None:
        """Apply the convention-parsed data to this commit.

        Parameters:
            convention: The convention to use.
        """
        self.convention.update(convention.parse_commit(self))

    def update_with_provider(
        self,
        provider: ProviderRefParser,
        parse_refs: bool = True,  # noqa: FBT001,FBT002
    ) -> None:
        """Apply the provider-parsed data to this commit.

        Parameters:
            provider: The provider to use.
            parse_refs: Whether to parse references for this provider.
        """
        # Set the commit URL based on the provider.
        # FIXME: hardcoded 'commits'
        if "commits" in provider.REF:
            self.url = provider.build_ref_url("commits", {"ref": self.hash})
        else:
            # Use default "commit" URL (could be wrong).
            self.url = f"{provider.url}/{provider.namespace}/{provider.project}/commit/{self.hash}"

        # Build commit text references from its subject and body.
        if parse_refs:
            for ref_type in provider.REF:
                self.text_refs[ref_type] = provider.get_refs(
                    ref_type,
                    "\n".join([self.subject, *self.body]),
                )

            if "issues" in self.text_refs:
                self.text_refs["issues_not_in_subject"] = []
                for issue in self.text_refs["issues"]:
                    if issue.ref not in self.subject:
                        self.text_refs["issues_not_in_subject"].append(issue)

    def _parse_trailers(self) -> None:
        last_blank_line = -1
        for index, line in enumerate(self.body):
            if not line:
                last_blank_line = index
        with suppress(ValueError):
            trailers = self._parse_trailers_block(self.body[last_blank_line + 1 :])
            if trailers:
                self.trailers.extend(trailers)
                self.body_without_trailers = self.body[:last_blank_line]

    def _parse_trailers_block(self, lines: list[str]) -> list[tuple[str, str]]:
        trailers = []
        for line in lines:
            title, value = line.split(": ", 1)
            trailers.append((title, value.strip()))
        return trailers  # Or raise `ValueError` due to split unpacking.


class CommitConvention(ABC):
    """A base class for a convention of commit messages."""

    TYPES: ClassVar[dict[str, str]]
    """The commit message types."""
    TYPE_REGEX: ClassVar[Pattern]
    """The commit message type regex."""
    BREAK_REGEX: ClassVar[Pattern]
    """The commit message breaking change regex."""
    DEFAULT_RENDER: ClassVar[list[str]]
    """The sections rendered by default."""

    @abstractmethod
    def parse_commit(self, commit: Commit) -> dict[str, str | bool]:
        """Parse the commit to extract information.

        Parameters:
            commit: The commit to parse.

        Returns:
            A dictionary containing the parsed data.
        """
        raise NotImplementedError

    @classmethod
    def _format_sections_help(cls) -> str:
        reversed_map = defaultdict(list)
        for section_type, section_title in cls.TYPES.items():
            reversed_map[section_title].append(section_type)
        default_sections = cls.DEFAULT_RENDER
        default = "- " + "\n- ".join(f"{', '.join(reversed_map[title])}: {title}" for title in default_sections)
        additional = "- " + "\n- ".join(
            f"{', '.join(types)}: {title}" for title, types in reversed_map.items() if title not in default_sections
        )
        return re.sub(
            r"\n *",
            "\n",
            f"""
            #### {cls.__name__[: -(len("Convention"))].strip()}

            *Default sections:*

            {default}

            *Additional sections:*

            {additional}
            """,
        )


class BasicConvention(CommitConvention):
    """Basic commit message convention."""

    TYPES: ClassVar[dict[str, str]] = {
        "add": "Added",
        "fix": "Fixed",
        "change": "Changed",
        "remove": "Removed",
        "merge": "Merged",
        "doc": "Documented",
    }
    """The commit message types."""

    TYPE_REGEX: ClassVar[Pattern] = re.compile(rf"^(?P<type>({'|'.join(TYPES.keys())}))", re.IGNORECASE)
    """The commit message type regex."""
    BREAK_REGEX: ClassVar[Pattern] = re.compile(
        r"^break(s|ing changes?)?[ :].+$",
        re.IGNORECASE | re.MULTILINE,
    )
    """The commit message breaking change regex."""
    DEFAULT_RENDER: ClassVar[list[str]] = [
        TYPES["add"],
        TYPES["fix"],
        TYPES["change"],
        TYPES["remove"],
    ]
    """The default sections to render."""

    def parse_commit(self, commit: Commit) -> dict[str, str | bool]:
        """Parse the commit to extract information.

        Parameters:
            commit: The commit to parse.

        Returns:
            A dictionary containing the parsed data.
        """
        commit_type = self.parse_type(commit.subject)
        message = "\n".join([commit.subject, *commit.body])
        is_major = self.is_major(message)
        is_minor = not is_major and self.is_minor(commit_type)
        is_patch = not any((is_major, is_minor))

        return {
            "type": commit_type,
            "is_major": is_major,
            "is_minor": is_minor,
            "is_patch": is_patch,
        }

    def parse_type(self, commit_subject: str) -> str:
        """Parse the type of the commit given its subject.

        Parameters:
            commit_subject: The commit message subject.

        Returns:
            The commit type.
        """
        type_match = self.TYPE_REGEX.match(commit_subject)
        if type_match:
            return self.TYPES.get(type_match.groupdict()["type"].lower(), "")
        return ""

    def is_minor(self, commit_type: str) -> bool:
        """Tell if this commit is worth a minor bump.

        Parameters:
            commit_type: The commit type.

        Returns:
            Whether it's a minor commit.
        """
        return commit_type == self.TYPES["add"]

    def is_major(self, commit_message: str) -> bool:
        """Tell if this commit is worth a major bump.

        Parameters:
            commit_message: The commit message.

        Returns:
            Whether it's a major commit.
        """
        return bool(self.BREAK_REGEX.search(commit_message))


class AngularConvention(CommitConvention):
    """Angular commit message convention."""

    TYPES: ClassVar[dict[str, str]] = {
        "build": "Build",
        "chore": "Chore",
        "ci": "Continuous Integration",
        "deps": "Dependencies",
        "doc": "Docs",
        "docs": "Docs",
        "feat": "Features",
        "fix": "Bug Fixes",
        "perf": "Performance Improvements",
        "ref": "Code Refactoring",
        "refactor": "Code Refactoring",
        "revert": "Reverts",
        "style": "Style",
        "test": "Tests",
        "tests": "Tests",
    }
    """The commit message types."""
    SUBJECT_REGEX: ClassVar[Pattern] = re.compile(
        rf"^(?P<type>({'|'.join(TYPES.keys())}))(?:\((?P<scope>.+)\))?: (?P<subject>.+)$",
    )
    """The commit message subject regex."""
    BREAK_REGEX: ClassVar[Pattern] = re.compile(
        r"^break(s|ing changes?)?[ :].+$",
        re.IGNORECASE | re.MULTILINE,
    )
    """The commit message breaking change regex."""
    DEFAULT_RENDER: ClassVar[list[str]] = [
        TYPES["feat"],
        TYPES["fix"],
        TYPES["revert"],
        TYPES["refactor"],
        TYPES["perf"],
    ]
    """The default sections to render."""

    def parse_commit(self, commit: Commit) -> dict[str, str | bool]:
        """Parse the commit to extract information.

        Parameters:
            commit: The commit to parse.

        Returns:
            A dictionary containing the parsed data.
        """
        subject = self.parse_subject(commit.subject)
        message = "\n".join([commit.subject, *commit.body])
        is_major = self.is_major(message)
        is_minor = not is_major and self.is_minor(subject["type"])
        is_patch = not any((is_major, is_minor))

        return {
            "type": subject["type"],
            "scope": subject["scope"],
            "subject": subject["subject"],
            "is_major": is_major,
            "is_minor": is_minor,
            "is_patch": is_patch,
        }

    def parse_subject(self, commit_subject: str) -> dict[str, str]:
        """Parse the subject of the commit (`<type>[(scope)]: Subject`).

        Parameters:
            commit_subject: The commit message subject.

        Returns:
            The parsed data.
        """
        subject_match = self.SUBJECT_REGEX.match(commit_subject)
        if subject_match:
            dct = subject_match.groupdict()
            dct["type"] = self.TYPES[dct["type"]]
            return dct
        return {"type": "", "scope": "", "subject": commit_subject}

    def is_minor(self, commit_type: str) -> bool:
        """Tell if this commit is worth a minor bump.

        Parameters:
            commit_type: The commit type.

        Returns:
            Whether it's a minor commit.
        """
        return commit_type == self.TYPES["feat"]

    def is_major(self, commit_message: str) -> bool:
        """Tell if this commit is worth a major bump.

        Parameters:
            commit_message: The commit message.

        Returns:
            Whether it's a major commit.
        """
        return bool(self.BREAK_REGEX.search(commit_message))


class ConventionalCommitConvention(AngularConvention):
    """Conventional commit message convention."""

    TYPES: ClassVar[dict[str, str]] = AngularConvention.TYPES
    """The commit message types."""
    DEFAULT_RENDER: ClassVar[list[str]] = AngularConvention.DEFAULT_RENDER
    """The default sections to render."""
    SUBJECT_REGEX: ClassVar[Pattern] = re.compile(
        rf"^(?P<type>({'|'.join(TYPES.keys())}))(?:\((?P<scope>.+)\))?(?P<breaking>!)?: (?P<subject>.+)$",
    )
    """The commit message subject regex."""

    def parse_commit(self, commit: Commit) -> dict[str, str | bool]:
        """Parse the commit to extract information.

        Parameters:
            commit: The commit to parse.

        Returns:
            A dictionary containing the parsed data.
        """
        subject = self.parse_subject(commit.subject)
        message = "\n".join([commit.subject, *commit.body])
        is_major = self.is_major(message) or subject.get("breaking") == "!"
        is_minor = not is_major and self.is_minor(subject["type"])
        is_patch = not any((is_major, is_minor))

        return {
            "type": subject["type"],
            "scope": subject["scope"],
            "subject": subject["subject"],
            "is_major": is_major,
            "is_minor": is_minor,
            "is_patch": is_patch,
        }


class LinuxConvention(BasicConvention):
    """Linux commit message convention."""

    def parse_type(self, commit_subject: str) -> str:
        """Parse the type of the commit given its subject.

        Parameters:
            commit_subject: The commit message subject.

        Returns:
            The commit type.
        """
        # Linux commit message are prefixed by "subsystem: " name.
        if ":" in commit_subject:
            commit_subject = commit_subject.split(":", 1)[1].lstrip()
        return super().parse_type(commit_subject)



"""git-changelog package.

Automatic Changelog generator using Jinja2 templates.
"""

from __future__ import annotations

from git_changelog._internal.build import Changelog, ConventionType, Section, Version, bump, parse_version
from git_changelog._internal.cli import (
    CONVENTIONS,
    DEFAULT_CHANGELOG_FILE,
    DEFAULT_CONFIG_FILES,
    DEFAULT_MARKER_LINE,
    DEFAULT_SETTINGS,
    DEFAULT_VERSION_REGEX,
    DEFAULT_VERSIONING,
    Templates,
    build,
    build_and_render,
    get_parser,
    get_release_notes,
    get_version,
    main,
    output_release_notes,
    parse_settings,
    providers,
    read_config,
    render,
)
from git_changelog._internal.commit import (
    AngularConvention,
    BasicConvention,
    Commit,
    CommitConvention,
    ConventionalCommitConvention,
    LinuxConvention,
)
from git_changelog._internal.providers import Bitbucket, GitHub, GitLab, ProviderRefParser, Ref, RefDef, RefRe
from git_changelog._internal.templates import (
    JINJA_ENV,
    TEMPLATES_PATH,
    configure_env,
    get_custom_template,
    get_template,
)
from git_changelog._internal.versioning import (
    ParsedVersion,
    PEP440Bumper,
    PEP440Strategy,
    PEP440Version,
    SemVerBumper,
    SemVerStrategy,
    SemVerVersion,
    VersionBumper,
    bump_pep440,
    bump_semver,
    parse_pep440,
    parse_semver,
    version_prefix,
)

__all__: list[str] = [
    "CONVENTIONS",
    "DEFAULT_CHANGELOG_FILE",
    "DEFAULT_CONFIG_FILES",
    "DEFAULT_MARKER_LINE",
    "DEFAULT_SETTINGS",
    "DEFAULT_VERSIONING",
    "DEFAULT_VERSION_REGEX",
    "JINJA_ENV",
    "TEMPLATES_PATH",
    "AngularConvention",
    "BasicConvention",
    "Bitbucket",
    "Changelog",
    "Commit",
    "CommitConvention",
    "ConventionType",
    "ConventionalCommitConvention",
    "GitHub",
    "GitLab",
    "LinuxConvention",
    "PEP440Bumper",
    "PEP440Strategy",
    "PEP440Version",
    "ParsedVersion",
    "ProviderRefParser",
    "Ref",
    "RefDef",
    "RefRe",
    "Section",
    "SemVerBumper",
    "SemVerStrategy",
    "SemVerVersion",
    "Templates",
    "Version",
    "VersionBumper",
    "build",
    "build_and_render",
    "bump",
    "bump_pep440",
    "bump_semver",
    "configure_env",
    "get_custom_template",
    "get_parser",
    "get_release_notes",
    "get_template",
    "get_version",
    "main",
    "output_release_notes",
    "parse_pep440",
    "parse_semver",
    "parse_settings",
    "parse_version",
    "providers",
    "read_config",
    "render",
    "version_prefix",
]



"""Tests for the `build` module."""

from __future__ import annotations

from time import sleep
from typing import TYPE_CHECKING, Literal

import pytest

from git_changelog import AngularConvention, BasicConvention, Changelog

if TYPE_CHECKING:
    from git_changelog import Version
    from tests.helpers import GitRepo


@pytest.mark.parametrize(
    ("versioning", "bump", "expected"),
    [
        ("semver", "auto", "0.0.1"),  # chore commit
        ("semver", "major", "1.0.0"),
        ("semver", "minor", "0.1.0"),
        ("semver", "patch", "0.0.1"),
        ("semver", "1.1.1", "1.1.1"),
        ("pep440", "auto", "0.0.1"),  # chore commit
        ("pep440", "major", "1.0.0"),
        ("pep440", "minor+dev", "0.1.0.dev0"),
        ("pep440", "micro+alpha+dev", "0.0.1a0.dev0"),
        ("pep440", "1.1.1", "1.1.1"),
    ],
)
def test_bump_with_semver_on_new_repo(
    repo: GitRepo,
    versioning: Literal["pep440", "semver"],
    bump: str,
    expected: str,
) -> None:
    """Bump to user specified version on new Git repo.

    Parameters:
        repo: GitRepo to a temporary repository.
        bump: The bump parameter value.
        expected: Expected version for the new changelog entry.
    """
    changelog = Changelog(
        repo.path,
        convention=AngularConvention,
        bump=bump,
        versioning=versioning,
        zerover=False,
    )
    assert len(changelog.versions_list) == 1
    assert changelog.versions_list[0].planned_tag == expected


@pytest.mark.parametrize("bump", ["auto", "major", "minor", "2.0.0"])
def test_no_bump_on_first_tag(repo: GitRepo, bump: str) -> None:
    """Ignore bump on new git repo without unreleased commits.

    Parameters:
        repo: GitRepo to a temporary repository.
        bump: The bump parameter value.
    """
    repo.tag("1.1.1")

    changelog = Changelog(repo.path, convention=AngularConvention, bump=bump)
    assert len(changelog.versions_list) == 1
    assert changelog.versions_list[0].tag == "1.1.1"


def test_one_release_branch_with_feat_branch(repo: GitRepo) -> None:
    r"""Test parsing and grouping commits to versions.

    Commit graph:
                   1.0.0
                     |
    main       A-B---D
                \   /
    feat         --C

    Expected:
    - 1.0.0: D B C A

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    commit_a = repo.first_hash
    repo.branch("develop")
    commit_b = repo.commit("fix: B")
    repo.checkout("develop")
    commit_c = repo.commit("feat: C")
    repo.checkout("main")
    commit_d = repo.merge("develop")
    repo.tag("1.0.0")

    changelog = Changelog(repo.path, convention=AngularConvention)

    assert len(changelog.versions_list) == 1
    _assert_version(
        changelog.versions_list[0],
        expected_tag="1.0.0",
        expected_prev_tag=None,
        expected_commits=[commit_d, commit_b, commit_c, commit_a],
    )


def test_one_release_branch_with_two_versions(repo: GitRepo) -> None:
    r"""Test parsing and grouping commits to versions.

    Commit graph:
                   1.1.0
               1.0.0 |
                 |   |
    main       A-B---D
                \   /
    feat         --C

    Expected:
    - 1.1.0: D C
    - 1.0.0: B A

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    commit_a = repo.first_hash
    repo.branch("develop")
    commit_b = repo.commit("fix: B")
    repo.tag("1.0.0")
    repo.checkout("develop")
    commit_c = repo.commit("feat: C")
    repo.checkout("main")
    commit_d = repo.merge("develop")
    repo.tag("1.1.0")

    changelog = Changelog(repo.path, convention=AngularConvention)

    assert len(changelog.versions_list) == 2
    _assert_version(
        changelog.versions_list[0],
        expected_tag="1.1.0",
        expected_prev_tag="1.0.0",
        expected_commits=[commit_d, commit_c],
    )
    _assert_version(
        changelog.versions_list[1],
        expected_tag="1.0.0",
        expected_prev_tag=None,
        expected_commits=[commit_b, commit_a],
    )


def test_two_release_branches(repo: GitRepo) -> None:
    r"""Test parsing and grouping commits to versions.

    Commit graph:
                   1.1.0
               1.0.0 |
                 |   |
    main       A-B---D-----G
                \     \   /
    develop      --C---E-F
                       |
                     2.0.0
    Expected:
    - Unreleased: G F
    - 2.0.0: E C
    - 1.1.0: D
    - 1.0.0: B A

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    commit_a = repo.first_hash
    repo.branch("develop")
    commit_b = repo.commit("fix: B")
    repo.tag("1.0.0")
    repo.checkout("develop")
    commit_c = repo.commit("feat: C")
    repo.checkout("main")
    commit_d = repo.commit("fix: C")
    repo.tag("1.1.0")
    repo.checkout("develop")
    # Git timestamp only has second precision,
    # so we delay to ensure git-log lists it before commit C.
    sleep(1)
    commit_e = repo.merge("main")
    repo.tag("2.0.0")
    commit_f = repo.commit("feat: F")
    repo.checkout("main")
    commit_g = repo.merge("develop")

    changelog = Changelog(repo.path, convention=AngularConvention)

    assert len(changelog.versions_list) == 4
    versions = iter(changelog.versions_list)
    _assert_version(
        next(versions),
        expected_tag="",
        expected_prev_tag="2.0.0",
        expected_commits=[commit_g, commit_f],
    )
    _assert_version(
        next(versions),
        expected_tag="2.0.0",
        expected_prev_tag="1.1.0",
        expected_commits=[commit_e, commit_c],
    )
    _assert_version(
        next(versions),
        expected_tag="1.1.0",
        expected_prev_tag="1.0.0",
        expected_commits=[commit_d],
    )
    _assert_version(
        next(versions),
        expected_tag="1.0.0",
        expected_prev_tag=None,
        expected_commits=[commit_b, commit_a],
    )


def _assert_version(
    version: Version,
    expected_tag: str,
    expected_prev_tag: str | None,
    expected_commits: list[str],
) -> None:
    assert expected_tag in (version.tag, version.planned_tag)
    if expected_prev_tag:
        assert version.previous_version is not None, f"Expected previous version '{expected_prev_tag}', but was None"
        assert version.previous_version.tag == expected_prev_tag
    else:
        assert version.previous_version is None
    hashes = [commit.hash for commit in version.commits]
    assert hashes == expected_commits


def test_no_remote_url(repo: GitRepo) -> None:
    """Test parsing and grouping commits to versions without a git remote.

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    repo.git("remote", "remove", "origin")
    commit_a = repo.first_hash
    repo.tag("1.0.0")

    changelog = Changelog(repo.path, convention=AngularConvention)

    assert len(changelog.versions_list) == 1
    _assert_version(
        changelog.versions_list[0],
        expected_tag="1.0.0",
        expected_prev_tag=None,
        expected_commits=[commit_a],
    )


def test_merge_into_unreleased(repo: GitRepo) -> None:
    r"""Test parsing and grouping commits to versions.

    Commit graph:
    main       A---C---E
                \ / \ /
    feat         B   D

    Expected:
    - Unreleased: E D C B A

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    commit_a = repo.first_hash
    repo.branch("feat/1")
    repo.checkout("feat/1")
    commit_b = repo.commit("feat: B")
    repo.checkout("main")
    commit_c = repo.merge("feat/1")
    repo.branch("feat/2")
    repo.checkout("feat/2")
    commit_d = repo.commit("feat: D")
    repo.checkout("main")
    commit_e = repo.merge("feat/2")

    changelog = Changelog(repo.path, convention=AngularConvention)

    assert len(changelog.versions_list) == 1
    version = changelog.versions_list[0]
    _assert_version(
        version,
        expected_tag="",
        expected_prev_tag=None,
        expected_commits=[commit_e, commit_c, commit_d, commit_a, commit_b],
    )


def test_build_changelog_with_pep440_versions(repo: GitRepo) -> None:
    """Test parsing and grouping commits to PEP440 versions.

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    repo.commit("feat: Feature")
    repo.tag("1.0.0")
    repo.commit("fix: Fix")
    repo.tag("1.0.0.post0")
    repo.commit("feat: Feat")
    changelog = Changelog(repo.path, convention=AngularConvention, versioning="pep440")
    assert len(changelog.versions_list) == 3
    assert changelog.versions_list[1].tag == "1.0.0.post0"


def test_ignore_nonsemver_tag(repo: GitRepo) -> None:
    """Test parsing and grouping commits to versions.

    Commit graph:
                 1.0.0
                   |
    main       A-B-C
                 |
               dummy

    Expected:
    - 1.0.0: C B A

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    commit_a = repo.first_hash
    commit_b = repo.commit("fix: B")
    repo.tag("dummy")
    commit_c = repo.commit("feat: C")
    repo.tag("1.0.0")

    changelog = Changelog(repo.path, convention=AngularConvention)

    assert len(changelog.versions_list) == 1
    _assert_version(
        changelog.versions_list[0],
        expected_tag="1.0.0",
        expected_prev_tag=None,
        expected_commits=[commit_c, commit_b, commit_a],
    )


def test_untyped_commits(repo: GitRepo) -> None:
    """Test capture of untyped (i.e. uncategorizable) commits.

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    commit_a = repo.first_hash
    commit_b = repo.commit(
        "this commit is untyped and therefore does not have a section!",
    )
    repo.tag("1.0.0")
    changelog = Changelog(repo.path, convention=AngularConvention)
    assert len(changelog.versions_list) == 1
    (version,) = changelog.versions_list
    assert len(version.sections_list) == 2
    typed_sections = changelog.versions_dict[version.tag].typed_sections
    assert len(typed_sections) == 1
    untyped = changelog.versions_dict[version.tag].untyped_section
    assert untyped is not None
    (typed,) = typed_sections
    assert len(untyped.commits) == 1
    assert len(typed.commits) == 1
    (untyped_commit,) = untyped.commits
    (typed_commit,) = typed.commits
    assert untyped_commit.hash == commit_b
    assert typed_commit.hash == commit_a


def test_include_all_false_excludes_untyped(repo: GitRepo) -> None:
    """Test that untyped commits are excluded from sections when include_all=False (default).

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    repo.commit("feat: Feature A")
    repo.commit("Random commit message without convention")
    repo.commit("fix: Fix B")
    repo.commit("Another random message")
    repo.tag("1.0.0")

    changelog = Changelog(repo.path, convention=AngularConvention, include_all=False)

    assert len(changelog.versions_list) == 1
    version = changelog.versions_list[0]

    # Empty string ("") should NOT be in sections when include_all=False.
    assert "" not in changelog.sections

    # Only conventional commits should be in the rendered sections.
    sections_dict = version.sections_dict
    # Should only have sections for typed commits (feat, fix, and chore from initial commit).
    assert "Features" in sections_dict
    assert "Bug Fixes" in sections_dict
    assert "Chore" in sections_dict

    # The untyped commits are captured but not rendered (empty string not in sections),
    # so they won't appear in the changelog output.


def test_include_all_true_includes_untyped(repo: GitRepo) -> None:
    """Test that untyped commits are included in sections when include_all=True.

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    repo.commit("feat: Feature A")
    commit_b = repo.commit("Random commit message without convention")
    repo.commit("fix: Fix C")
    commit_d = repo.commit("Another random message")
    repo.tag("1.0.0")

    changelog = Changelog(repo.path, convention=AngularConvention, include_all=True)

    assert len(changelog.versions_list) == 1
    version = changelog.versions_list[0]

    # Empty string ("") SHOULD be in sections when include_all=True.
    assert "" in changelog.sections

    # All sections should be present.
    sections_dict = version.sections_dict
    assert "Features" in sections_dict
    assert "Bug Fixes" in sections_dict
    assert "Chore" in sections_dict  # From initial commit
    assert "" in sections_dict  # Untyped section

    # Check that untyped section contains the non-conventional commits.
    untyped_section = sections_dict[""]
    assert len(untyped_section.commits) == 2
    untyped_hashes = [commit.hash for commit in untyped_section.commits]
    assert commit_b in untyped_hashes
    assert commit_d in untyped_hashes


def test_include_all_with_sections_parameter(repo: GitRepo) -> None:
    """Test that include_all works correctly with explicit sections parameter.

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    repo.commit("feat: Feature A")
    commit_b = repo.commit("Random commit message")
    repo.commit("fix: Fix C")
    repo.tag("1.0.0")

    # Specify only feat and fix sections, but with include_all=True.
    changelog = Changelog(
        repo.path,
        convention=AngularConvention,
        sections=["feat", "fix"],
        include_all=True,
    )

    # Empty string should be added at the end.
    assert "" in changelog.sections
    # The specified sections should be present.
    assert "Features" in [s for s in changelog.sections if s]
    assert "Bug Fixes" in [s for s in changelog.sections if s]

    version = changelog.versions_list[0]
    sections_dict = version.sections_dict

    # Untyped section should be present.
    assert "" in sections_dict
    untyped_section = sections_dict[""]
    assert len(untyped_section.commits) == 1
    assert untyped_section.commits[0].hash == commit_b


def test_sections_all_expands_to_all_types(repo: GitRepo) -> None:
    """Test that :all: expands to all available section types for the convention.

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    repo.tag("1.0.0")
    changelog = Changelog(repo.path, convention=AngularConvention, sections=[":all:"])

    # AngularConvention.TYPES contains all available section types.
    expected_sections = list(dict.fromkeys(AngularConvention.TYPES.values()))
    assert changelog.sections == expected_sections


def test_sections_all_with_basic_convention(repo: GitRepo) -> None:
    """Test that :all: works with BasicConvention.

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    repo.tag("1.0.0")
    changelog = Changelog(repo.path, convention=BasicConvention, sections=[":all:"])

    # BasicConvention.TYPES contains all available section types.
    expected_sections = list(BasicConvention.TYPES.values())
    assert changelog.sections == expected_sections


def test_sections_explicit_list_still_works(repo: GitRepo) -> None:
    """Test that explicit section list still works as before.

    Parameters:
        repo: GitRepo to a temporary repository.
    """
    repo.tag("1.0.0")
    changelog = Changelog(
        repo.path,
        convention=AngularConvention,
        sections=["feat", "fix"],
    )

    # Should map to full section names.
    assert changelog.sections == ["Features", "Bug Fixes"]



# Why does this file exist, and why not put this in `__main__`?
#
# You might be tempted to import things from `__main__` later,
# but that will cause problems: the code will get executed twice:
#
# - When you run `python -m git_changelog` python will execute
#   `__main__.py` as a script. That means there won't be any
#   `git_changelog.__main__` in `sys.modules`.
# - When you import `__main__` it will get executed again (as a module) because
#   there's no `git_changelog.__main__` in `sys.modules`.

from __future__ import annotations

import argparse
import re
import sys
import warnings
from importlib import metadata
from pathlib import Path
from re import Pattern
from typing import TYPE_CHECKING, Any, Literal, TextIO

from jinja2.exceptions import TemplateNotFound
from platformdirs import user_config_dir

from git_changelog._internal import debug, templates
from git_changelog._internal.build import Changelog, Version
from git_changelog._internal.commit import (
    AngularConvention,
    BasicConvention,
    CommitConvention,
    ConventionalCommitConvention,
    LinuxConvention,
)
from git_changelog._internal.providers import Bitbucket, GitHub, GitLab, ProviderRefParser
from git_changelog._internal.versioning import bump_pep440, bump_semver

# YORE: EOL 3.10: Replace block with line 2.
if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

if TYPE_CHECKING:
    from collections.abc import Sequence

_MARKER_LINE_PREPEND = ":prepend:"

DEFAULT_VERSIONING = "semver"
"""Default versioning strategy."""
DEFAULT_VERSION_REGEX = r"^## \[(?P<version>v?[^\]]+)"
"""Default version regex for finding versions."""
DEFAULT_MARKER_LINE = "<!-- insertion marker -->"
"""Default marker line for finding insertion points."""
DEFAULT_CHANGELOG_FILE = "CHANGELOG.md"
"""Default changelog file name."""
CONVENTIONS = sorted(Changelog.CONVENTION)
"""Available commit message conventions."""
DEFAULT_CONFIG_FILES = [
    "pyproject.toml",
    ".git-changelog.toml",
    "config/git-changelog.toml",
    ".config/git-changelog.toml",
    str(Path(user_config_dir()) / "git-changelog.toml"),
]
"""Default configuration files read by git-changelog."""

_DEFAULT_DEBIAN_MARKER_LINE = _MARKER_LINE_PREPEND
_DEFAULT_DEBIAN_VERSION_REGEX = r"^[\w-]+ \((?:\d+:)?(?P<version>[^-+~)]+)(?!.*UNRELEASED)"

DEFAULT_SETTINGS: dict[str, Any] = {
    "bump": None,
    # YORE: Bump 3: Remove line.
    "bump_latest": None,
    "convention": "basic",
    "filter_commits": None,
    "in_place": False,
    "include_all": False,
    "input": DEFAULT_CHANGELOG_FILE,
    "marker_line": None,
    "omit_empty_versions": False,
    "output": sys.stdout,
    "parse_refs": False,
    "parse_trailers": False,
    "provider": None,
    "release_notes": False,
    "bumped_version": False,
    "repository": ".",
    "sections": None,
    "template": "keepachangelog",
    "jinja_context": {},
    "version_regex": None,
    "versioning": DEFAULT_VERSIONING,
    "zerover": True,
}
"""Default settings for the CLI."""


class Templates(tuple):  # noqa: SLOT001
    """Helper to pick a template on the command line."""

    def __contains__(self, item: object) -> bool:
        """Check if the template is in the list."""
        if isinstance(item, str):
            return item.startswith("path:") or super().__contains__(item)
        return False


def get_version() -> str:
    """Return the current `git-changelog` version.

    Returns:
        The current `git-changelog` version.
    """
    try:
        return metadata.version("git-changelog")
    except metadata.PackageNotFoundError:
        return "0.0.0"


def _comma_separated_list(value: str) -> list[str]:
    return value.split(",")


class _ParseDictAction(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,  # noqa: ARG002
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | Sequence[Any] | None = None,  # noqa: ARG002
    ):
        attribute = getattr(namespace, self.dest)
        if not isinstance(attribute, dict):
            setattr(namespace, self.dest, {})
        if isinstance(values, str):
            key, value = values.split("=", 1)
            getattr(namespace, self.dest)[key] = value


providers: dict[str, type[ProviderRefParser]] = {
    "github": GitHub,
    "gitlab": GitLab,
    "bitbucket": Bitbucket,
}
"""Available version control providers."""


class _DebugInfo(argparse.Action):
    def __init__(self, nargs: int | str | None = 0, **kwargs: Any) -> None:
        super().__init__(nargs=nargs, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ARG002
        debug._print_debug_info()
        sys.exit(0)


def get_parser() -> argparse.ArgumentParser:
    """Return the CLI argument parser.

    Returns:
        An argparse parser.
    """
    parser = argparse.ArgumentParser(
        add_help=False,
        prog="git-changelog",
        description=re.sub(
            r"\n *",
            "\n",
            f"""
            Automatic Changelog generator using Jinja2 templates.

            This tool parses your commit messages to extract useful data
            that is then rendered using Jinja2 templates, for example to
            a changelog file formatted in Markdown.

            Each Git tag will be treated as a version of your project.
            Each version contains a set of commits, and will be an entry
            in your changelog. Commits in each version will be grouped
            by sections, depending on the commit convention you follow.

            ### Conventions

            {BasicConvention._format_sections_help()}
            {AngularConvention._format_sections_help()}
            {ConventionalCommitConvention._format_sections_help()}
            {LinuxConvention._format_sections_help()}
            """,
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "repository",
        metavar="REPOSITORY",
        nargs="?",
        help="The repository path, relative or absolute. Default: current working directory.",
    )

    parser.add_argument(
        "--config-file",
        metavar="PATH",
        nargs="*",
        help="Configuration file(s).",
    )

    # YORE: Bump 3: Remove block.
    parser.add_argument(
        "-b",
        "--bump-latest",
        action="store_true",
        dest="bump_latest",
        help="Deprecated, use --bump=auto instead. "
        "Guess the new latest version by bumping the previous one based on the set of unreleased commits. "
        "For example, if a commit contains breaking changes, bump the major number (or the minor number for 0.x versions). "
        "Else if there are new features, bump the minor number. Else just bump the patch number. "
        "Default: unset (false).",
    )

    parser.add_argument(
        "--bumped-version",
        action="store_true",
        dest="bumped_version",
        help="Show bumped version and exit.",
    )
    parser.add_argument(
        "-B",
        "--bump",
        action="store",
        dest="bump",
        metavar="VERSION",
        help="Specify the bump from latest version for the set of unreleased commits. "
        "Can be one of `auto`, `major`, `minor`, `patch` or a valid SemVer version (eg. 1.2.3). "
        "For both SemVer and PEP 440 versioning schemes (see -n), `auto` will bump the major number "
        "if a commit contains breaking changes (or the minor number for 0.x versions, see -Z), "
        "else the minor number if there are new features, else the patch number. "
        "Default: unset (false).",
    )
    parser.add_argument(
        "-n",
        "--versioning",
        action="store",
        dest="versioning",
        metavar="SCHEME",
        help="Versioning scheme to use when bumping and comparing versions. "
        "The selected scheme will impact the values accepted by the `--bump` option. "
        "Supported: `pep440`, `semver`. PEP 440 provides the following bump strategies: `auto`, "
        f"`{'`, `'.join(part for part in bump_pep440.strategies if '+' not in part)}`. "
        "Values `auto`, `major`, `minor`, `micro` can be suffixed with one of `+alpha`, `+beta`, `+candidate`, "
        "and/or `+dev`. Values `alpha`, `beta` and `candidate` can be suffixed with `+dev`. "
        "Examples: `auto+alpha`, `major+beta+dev`, `micro+dev`, `candidate+dev`, etc.. "
        "SemVer provides the following bump strategies: `auto`, "
        f"`{'`, `'.join(bump_semver.strategies)}`. "
        "See the docs for more information. Default: unset (`semver`).",
    )
    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit.",
    )
    parser.add_argument(
        "-i",
        "--in-place",
        action="store_true",
        dest="in_place",
        help="Insert new entries (versions missing from changelog) in-place. "
        "An output file must be specified. With custom templates, "
        "you can pass two additional Parameters: `--version-regex` and `--marker-line`. "
        "When writing in-place, an `in_place` variable "
        "will be injected in the Jinja context, "
        "allowing to adapt the generated contents "
        "(for example to skip changelog headers or footers). Default: unset (false).",
    )
    parser.add_argument(
        "-g",
        "--version-regex",
        action="store",
        metavar="REGEX",
        dest="version_regex",
        help="A regular expression to match versions in the existing changelog "
        "(used to find the latest release) when writing in-place. "
        "The regular expression must be a Python regex with a `version` named group. "
        f"Default: `{DEFAULT_VERSION_REGEX}`.",
    )

    parser.add_argument(
        "-m",
        "--marker-line",
        action="store",
        metavar="MARKER",
        dest="marker_line",
        help="A marker line at which to insert new entries "
        "(versions missing from changelog). "
        "If two marker lines are present in the changelog, "
        "the contents between those two lines will be overwritten "
        "(useful to update an 'Unreleased' entry for example). "
        f"Default: `{DEFAULT_MARKER_LINE}`.",
    )
    parser.add_argument(
        "-o",
        "--output",
        action="store",
        metavar="FILE",
        dest="output",
        help="Output to given file. Default: standard output.",
    )
    parser.add_argument(
        "-p",
        "--provider",
        metavar="PROVIDER",
        dest="provider",
        choices=providers.keys(),
        help="Explicitly specify the repository provider. Default: unset.",
    )
    parser.add_argument(
        "-r",
        "--parse-refs",
        action="store_true",
        dest="parse_refs",
        help="Parse provider-specific references in commit messages (GitHub/GitLab/Bitbucket "
        "issues, PRs, etc.). Default: unset (false).",
    )
    parser.add_argument(
        "-R",
        "--release-notes",
        action="store_true",
        dest="release_notes",
        help="Output release notes to stdout based on the last entry in the changelog. Default: unset (false).",
    )
    parser.add_argument(
        "-I",
        "--input",
        metavar="FILE",
        dest="input",
        help=f"Read from given file when creating release notes. Default: `{DEFAULT_CHANGELOG_FILE}`.",
    )
    parser.add_argument(
        "-c",
        "--convention",
        "--commit-style",
        "--style",
        metavar="CONVENTION",
        choices=CONVENTIONS,
        dest="convention",
        help=f"The commit convention to match against. Default: `{DEFAULT_SETTINGS['convention']}`.",
    )
    parser.add_argument(
        "-s",
        "--sections",
        action="store",
        type=_comma_separated_list,
        metavar="SECTIONS",
        dest="sections",
        help="A comma-separated list of sections to render. "
        "Use `:all:` to include all available sections for the selected convention. "
        "See the available sections for each supported convention in the description. "
        "Default: unset (None).",
    )
    parser.add_argument(
        "-a",
        "--include-all",
        action="store_true",
        dest="include_all",
        help="Include all commits, even those without a recognized type. "
        "These commits will be rendered in a 'Misc' section at the end of each version. "
        "Default: unset (false).",
    )
    parser.add_argument(
        "-t",
        "--template",
        choices=Templates(("angular", "keepachangelog", "debian")),
        metavar="TEMPLATE",
        dest="template",
        help="The Jinja2 template to use. Prefix it with `path:` to specify the path "
        "to a Jinja templated file. "
        f"Default: `{DEFAULT_SETTINGS['template']}`.",
    )
    parser.add_argument(
        "-T",
        "--trailers",
        "--git-trailers",
        action="store_true",
        dest="parse_trailers",
        help="Parse Git trailers in the commit message. "
        "See https://git-scm.com/docs/git-interpret-trailers. Default: unset (false).",
    )
    parser.add_argument(
        "-E",
        "--omit-empty-versions",
        action="store_true",
        dest="omit_empty_versions",
        help="Omit empty versions from the output. Default: unset (false).",
    )
    parser.add_argument(
        "-Z",
        "--no-zerover",
        action="store_false",
        dest="zerover",
        help="By default, breaking changes on a 0.x don't bump the major version, maintaining it at 0. "
        "With this option, a breaking change will bump a 0.x version to 1.0.",
    )
    parser.add_argument(
        "-F",
        "--filter-commits",
        action="store",
        metavar="RANGE",
        dest="filter_commits",
        help="The Git revision-range filter to use (e.g. `v1.2.0..`). Default: no filter.",
    )
    parser.add_argument(
        "-j",
        "--jinja-context",
        action=_ParseDictAction,
        metavar="KEY=VALUE",
        dest="jinja_context",
        help="Pass additional key/value pairs to the template. Option can be used multiple times. "
        "The key/value pairs are accessible as 'jinja_context' in the template.",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {debug._get_version()}",
        help="Show the current version of the program and exit.",
    )
    parser.add_argument("--debug-info", action=_DebugInfo, help="Print debug information.")

    return parser


def _latest(lines: list[str], regex: Pattern) -> tuple[str | None, int]:
    for idx, line in enumerate(lines):
        match = regex.search(line)
        if match:
            return match.groupdict()["version"], idx
    return None, 0


def _unreleased(versions: list[Version], last_release: str) -> list[Version]:
    for index, version in enumerate(versions):
        if version.tag == last_release:
            return versions[:index]
    return versions


def read_config(
    config_file: Sequence[str | Path] | str | Path | None = DEFAULT_CONFIG_FILES,
) -> dict:
    """Find config files and initialize settings with the one of highest priority.

    Parameters:
        config_file: A path or list of paths to configuration file(s); or `None` to
            disable config file settings. Default: a list of paths given by
            [`git_changelog.DEFAULT_CONFIG_FILES`][].

    Returns:
        A settings dictionary. Default settings if no config file is found or `config_file` is `None`.

    """
    project_config = DEFAULT_SETTINGS.copy()
    if config_file is None:  # Unset config file.
        return project_config

    for filename in config_file if isinstance(config_file, (list, tuple)) else [config_file]:
        _path = Path(filename)  # ty:ignore[invalid-argument-type]

        if not _path.exists():
            continue

        with _path.open("rb") as file:
            new_settings = tomllib.load(file)
        if _path.name == "pyproject.toml":
            new_settings = new_settings.get("tool", {}).get("git-changelog", {}) or new_settings.get(
                "tool.git-changelog",
                {},
            )

            if not new_settings:  # pyproject.toml did not have a git-changelog section.
                continue

        # Settings can have hyphens like in the CLI.
        new_settings = {key.replace("-", "_"): value for key, value in new_settings.items()}

        # YORE: Bump 3: Remove block.
        if "bump_latest" in new_settings:
            _opt_value = new_settings["bump_latest"]
            _suggestion = (
                "remove it from the config file" if not _opt_value else "set `bump = 'auto'` in the config file instead"
            )
            warnings.warn(
                f"`bump-latest = {str(_opt_value).lower()}` option found "
                f"in config file ({_path.absolute()}). This option will be removed in the future. "
                f"To achieve the same result, please {_suggestion}.",
                FutureWarning,
                stacklevel=1,
            )

        # Massage found values to meet expectations.
        # Parse sections.
        if "sections" in new_settings:
            # Remove "sections" from dict, only restore if the list is valid.
            sections = new_settings.pop("sections", None)
            if isinstance(sections, str):
                sections = sections.split(",")

            sections = [s.strip() for s in sections if isinstance(s, str) and s.strip()]

            if sections:  # TOML doesn't store null/nil.
                new_settings["sections"] = sections

        project_config.update(new_settings)
        break

    return project_config


def parse_settings(args: list[str] | None = None) -> dict:
    """Parse arguments and config files to build the final settings set.

    Parameters:
        args: Arguments passed from the command line.

    Returns:
        A dictionary with the final settings.
    """
    parser = get_parser()
    opts = vars(parser.parse_args(args=args))

    # Determine which arguments were explicitly set with the CLI
    sentinel = object()
    sentinel_ns = argparse.Namespace(**dict.fromkeys(opts, sentinel))
    explicit_opts_dict = {
        key: opts.get(key, None)
        for key, value in vars(parser.parse_args(namespace=sentinel_ns, args=args)).items()
        if value is not sentinel
    }

    config_file = explicit_opts_dict.pop("config_file", DEFAULT_CONFIG_FILES)
    if str(config_file).strip().lower() in ("no", "none", "off", "false", "0", ""):
        config_file = None
    elif str(config_file).strip().lower() in ("yes", "default", "on", "true", "1"):
        config_file = DEFAULT_CONFIG_FILES

    jinja_context = explicit_opts_dict.pop("jinja_context", {})

    settings = read_config(config_file)

    # CLI arguments override the config file settings
    settings.update(explicit_opts_dict)

    # Merge jinja context, CLI values override config file values.
    settings["jinja_context"].update(jinja_context)

    # YORE: Bump 3: Remove block.
    if "bump_latest" in explicit_opts_dict:
        warnings.warn("`--bump-latest` is deprecated in favor of `--bump=auto`", FutureWarning, stacklevel=1)

    return settings


def build(
    repository: str,
    convention: str | CommitConvention,
    parse_refs: bool = False,  # noqa: FBT001,FBT002
    parse_trailers: bool = False,  # noqa: FBT001,FBT002
    sections: list[str] | None = None,
    # YORE: Bump 3: Remove line.
    bump_latest: bool = False,  # noqa: FBT001,FBT002
    omit_empty_versions: bool = False,  # noqa: FBT001,FBT002
    provider: str | None = None,
    bump: str | None = None,
    zerover: bool = True,  # noqa: FBT001,FBT002
    filter_commits: str | None = None,
    include_all: bool = False,  # noqa: FBT001,FBT002
    versioning: Literal["pep440", "semver"] = "semver",
    **kwargs: Any,  # noqa: ARG001
) -> Changelog:
    """Build a changelog.

    Parameters:
        repository: Path to a local repository.
        convention: Name of a commit message style/convention.
        parse_refs: Whether to parse provider-specific references (GitHub/GitLab issues, PRs, etc.).
        parse_trailers: Whether to parse Git trailers.
        sections: Sections to render (features, bug fixes, etc.).
        bump_latest: Deprecated, use --bump=auto instead.
            Whether to try and bump the latest version to guess the new one.
        omit_empty_versions: Whether to omit empty versions from the output.
        provider: Provider class used by this repository.
        bump: Whether to try and bump to a given version.
        zerover: Keep major version at zero, even for breaking changes.
        filter_commits: The Git revision-range used to filter commits in git-log.
        include_all: Whether to include commits without a recognized type.
        versioning: Versioning scheme to use when grouping commits and bumping versions.
        **kwargs: Swallowing kwargs to allow passing all settings at once.

    Raises:
        ValueError: When some arguments are incompatible or missing.

    Returns:
        The built changelog.
    """
    # YORE: Bump 3: Remove block.
    if bump_latest:
        warnings.warn("`bump_latest=True` is deprecated in favor of `bump='auto'`", DeprecationWarning, stacklevel=1)
        if bump is None:
            bump = "auto"

    # Build data.
    changelog = Changelog(
        repository,
        provider=providers[provider] if provider else None,
        convention=convention,
        parse_provider_refs=parse_refs,
        parse_trailers=parse_trailers,
        sections=sections,
        bump=bump,
        zerover=zerover,
        filter_commits=filter_commits,
        include_all=include_all,
        versioning=versioning,
    )

    # Remove empty versions from changelog data.
    if omit_empty_versions:
        section_set = set(changelog.sections)
        empty_versions = [
            version for version in changelog.versions_list if section_set.isdisjoint(version.sections_dict.keys())
        ]
        for version in empty_versions:
            changelog.versions_list.remove(version)
            changelog.versions_dict.pop(version.tag)

    return changelog


def render(
    changelog: Changelog,
    template: str,
    in_place: bool = False,  # noqa: FBT001,FBT002
    output: str | TextIO | None = None,
    version_regex: str | None = None,
    marker_line: str | None = None,
    # YORE: Bump 3: Remove line.
    bump_latest: bool = False,  # noqa: FBT001,FBT002
    bump: str | None = None,
    jinja_context: dict[str, Any] | None = None,
) -> str:
    """Render a changelog.

    This function updates the specified output file (side-effect) or writes to stdout.

    Parameters:
        changelog: The changelog to render.
        template: Name of a builtin template, or path to a custom template (prefixed with `path:`).
        in_place: Whether to update the changelog in-place.
        output: Output/changelog file.
        version_regex: Regular expression to match versions in an existing changelog file.
        marker_line: Marker line used to insert contents in an existing changelog.
        bump_latest: Deprecated, use --bump=auto instead.
            Whether to try and bump the latest version to guess the new one.
        bump: Whether to try and bump to a given version.
        jinja_context: Key/value pairs passed to the Jinja template.

    Raises:
        ValueError: When some arguments are incompatible or missing.

    Returns:
        The rendered changelog.
    """
    # Get template.
    if template.startswith("path:"):
        path = template.replace("path:", "", 1)
        try:
            jinja_template = templates.get_custom_template(path)
        except TemplateNotFound as error:
            raise ValueError(f"No such file: {path}") from error
    else:
        jinja_template = templates.get_template(template)

    if output is None:
        output = sys.stdout

    # Handle misconfiguration early.
    if in_place and output is sys.stdout:
        raise ValueError("Cannot write in-place to stdout")

    # YORE: Bump 3: Remove block.
    if bump_latest:
        warnings.warn("`bump_latest=True` is deprecated in favor of `bump='auto'`", DeprecationWarning, stacklevel=1)
        if bump is None:
            bump = "auto"

    # Render new entries in-place.
    if in_place:
        # Read current changelog lines.
        with open(output, encoding="utf8") as changelog_file:  # ty:ignore[invalid-argument-type]
            lines = changelog_file.read().splitlines()

        # Prepare version regex and marker line.
        if version_regex is None:
            version_regex = _DEFAULT_DEBIAN_VERSION_REGEX if template == "debian" else DEFAULT_VERSION_REGEX
        if marker_line is None:
            marker_line = _DEFAULT_DEBIAN_MARKER_LINE if template == "debian" else DEFAULT_MARKER_LINE

        # Only keep new entries (missing from changelog).
        last_released, last_released_line = _latest(lines, re.compile(version_regex))
        if last_released:
            # Check if the latest version is already in the changelog.
            if last_released in [
                changelog.versions_list[0].tag,
                changelog.versions_list[0].planned_tag,
            ]:
                raise ValueError(f"Version {last_released} already in changelog")
            changelog.versions_list = _unreleased(
                changelog.versions_list,
                last_released,
            )

        # Render new entries.
        rendered = (
            jinja_template.render(
                changelog=changelog,
                jinja_context=jinja_context,
                in_place=True,
            ).rstrip("\n")
            + "\n"
        )

        # Find marker line(s) in current changelog.
        if marker_line == ":prepend:":
            # With prepend mode, take account of last_released_line to replace unreleased changelog.
            lines[0:last_released_line] = [rendered]
        else:
            marker = lines.index(marker_line)
            try:
                marker2 = lines[marker + 1 :].index(marker_line)
            except ValueError:
                # Apply new entries at marker line.
                lines[marker] = rendered
            else:
                # Apply new entries between marker lines.
                lines[marker : marker + marker2 + 2] = [rendered]

        # Write back updated changelog lines.
        with open(output, "w", encoding="utf8") as changelog_file:  # ty:ignore[no-matching-overload]
            changelog_file.write("\n".join(lines).rstrip("\n") + "\n")

    # Overwrite output file.
    else:
        rendered = jinja_template.render(changelog=changelog, jinja_context=jinja_context)

        # Write result in specified output.
        if output is sys.stdout:
            sys.stdout.write(rendered)
        else:
            with open(output, "w", encoding="utf8") as stream:  # ty:ignore[no-matching-overload]
                stream.write(rendered)

    return rendered


def build_and_render(
    repository: str,
    template: str,
    convention: str | CommitConvention,
    parse_refs: bool = False,  # noqa: FBT001,FBT002
    parse_trailers: bool = False,  # noqa: FBT001,FBT002
    sections: list[str] | None = None,
    in_place: bool = False,  # noqa: FBT001,FBT002
    output: str | TextIO | None = None,
    version_regex: str | None = None,
    marker_line: str | None = None,
    # YORE: Bump 3: Remove line.
    bump_latest: bool = False,  # noqa: FBT001,FBT002
    omit_empty_versions: bool = False,  # noqa: FBT001,FBT002
    provider: str | None = None,
    bump: str | None = None,
    zerover: bool = True,  # noqa: FBT001,FBT002
    filter_commits: str | None = None,
    jinja_context: dict[str, Any] | None = None,
    versioning: Literal["pep440", "semver"] = "semver",
    include_all: bool = False,  # noqa: FBT001,FBT002
    **kwargs: Any,  # noqa: ARG001
) -> tuple[Changelog, str]:
    """Build a changelog and render it.

    This function returns the changelog instance and the rendered contents,
    but also updates the specified output file (side-effect) or writes to stdout.

    Parameters:
        repository: Path to a local repository.
        template: Name of a builtin template, or path to a custom template (prefixed with `path:`).
        convention: Name of a commit message style/convention.
        parse_refs: Whether to parse provider-specific references (GitHub/GitLab issues, PRs, etc.).
        parse_trailers: Whether to parse Git trailers.
        sections: Sections to render (features, bug fixes, etc.).
        in_place: Whether to update the changelog in-place.
        output: Output/changelog file.
        version_regex: Regular expression to match versions in an existing changelog file.
        marker_line: Marker line used to insert contents in an existing changelog.
        bump_latest: Deprecated, use --bump=auto instead.
            Whether to try and bump the latest version to guess the new one.
        omit_empty_versions: Whether to omit empty versions from the output.
        provider: Provider class used by this repository.
        bump: Whether to try and bump to a given version.
        zerover: Keep major version at zero, even for breaking changes.
        filter_commits: The Git revision-range used to filter commits in git-log.
        include_all: Whether to include commits without a recognized type.
        jinja_context: Key/value pairs passed to the Jinja template.
        versioning: Versioning scheme to use when grouping commits and bumping versions.
        **kwargs: Swallowing kwargs to allow passing all settings at once.

    Raises:
        ValueError: When some arguments are incompatible or missing.

    Returns:
        The built changelog and the rendered contents.
    """
    changelog = build(
        repository=repository,
        convention=convention,
        parse_refs=parse_refs,
        parse_trailers=parse_trailers,
        sections=sections,
        bump_latest=bump_latest,
        omit_empty_versions=omit_empty_versions,
        provider=provider,
        bump=bump,
        zerover=zerover,
        filter_commits=filter_commits,
        include_all=include_all,
        versioning=versioning,
    )
    rendered = render(
        changelog=changelog,
        template=template,
        in_place=in_place,
        output=output,
        version_regex=version_regex,
        marker_line=marker_line,
        bump_latest=bump_latest,
        bump=bump,
        jinja_context=jinja_context,
    )
    return changelog, rendered


def get_release_notes(
    input_file: str | Path = "CHANGELOG.md",
    version_regex: str = DEFAULT_VERSION_REGEX,
    marker_line: str = DEFAULT_MARKER_LINE,
) -> str:
    """Get release notes from existing changelog.

    This will return the latest entry in the changelog.

    Parameters:
        input_file: The changelog to read from.
        version_regex: A regular expression to match version entries.
        marker_line: The insertion marker line in the changelog.

    Returns:
        The latest changelog entry.
    """
    release_notes = []
    found_version = False
    with open(input_file, encoding="utf8") as changelog:
        for line in changelog:
            if re.search(version_regex, line):
                if found_version:
                    break
                found_version = True
                release_notes.append(line.rstrip("\n"))
            elif found_version:
                release_notes.append(line.rstrip("\n"))
    result = "\n".join(release_notes).strip()
    if result.endswith(marker_line):
        result = result[: -len(marker_line)].strip()
    return result


def output_release_notes(
    input_file: str = "CHANGELOG.md",
    version_regex: str | None = None,
    marker_line: str | None = None,
    output_file: str | TextIO | None = None,
) -> None:
    """Print release notes from existing changelog.

    This will print the latest entry in the changelog.

    Parameters:
        input_file: The changelog to read from.
        version_regex: A regular expression to match version entries.
        marker_line: The insertion marker line in the changelog.
        output_file: Where to print/write the release notes.
    """
    version_regex = version_regex or DEFAULT_VERSION_REGEX
    marker_line = marker_line or DEFAULT_MARKER_LINE
    output_file = output_file or sys.stdout
    release_notes = get_release_notes(input_file, version_regex, marker_line)
    try:
        output_file.write(release_notes)  # ty:ignore[unresolved-attribute]
    except AttributeError:
        with open(output_file, "w") as file:  # ty:ignore[no-matching-overload]
            file.write(release_notes)


def main(args: list[str] | None = None) -> int:
    """Run the main program.

    This function is executed when you type `git-changelog` or `python -m git_changelog`.

    Parameters:
        args: Arguments passed from the command line.

    Returns:
        An exit code.
    """
    settings = parse_settings(args)

    if settings.pop("release_notes"):
        output_release_notes(
            input_file=settings["input"],
            version_regex=settings["version_regex"],
            marker_line=settings["marker_line"],
            output_file=None,  # force writing to stdout
        )
        return 0

    if settings.pop("bumped_version"):
        settings["bump"] = settings.get("bump") or "auto"
        changelog = build(**settings)
        if not changelog.versions_list:
            print("git-changelog: No version found in the repository.", file=sys.stderr)
            return 1
        print(changelog.versions_list[0].planned_tag)
        return 0

    try:
        build_and_render(**settings)
    except ValueError as error:
        print(f"git-changelog: {error}", file=sys.stderr)
        return 1

    return 0



"""Tests for the CLI."""

# IMPORTANT: Do not call `git_changelog.main()`
# without passing a config file path, otherwise
# it will use its own config file and possibly modify
# the CHANGELOG!

from __future__ import annotations

import os
import re
import sys
from textwrap import dedent
from typing import TYPE_CHECKING, Any

import pytest
import tomli_w

from git_changelog import CONVENTIONS, DEFAULT_SETTINGS, get_version, main, parse_settings, read_config
from git_changelog._internal import debug
from git_changelog._internal.cli import _DEFAULT_DEBIAN_VERSION_REGEX

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from tests.helpers import GitRepo


# YORE: EOL 3.10: Replace block with line 2.
if sys.version_info >= (3, 11):
    from contextlib import chdir
else:
    from contextlib import contextmanager

    @contextmanager
    def chdir(path: str) -> Iterator[None]:
        old_wd = os.getcwd()  # noqa: PTH109
        os.chdir(path)
        try:
            yield
        finally:
            os.chdir(old_wd)


# IMPORTANT: See top module comment.
def test_main(tmp_path: Path) -> None:
    """Basic CLI test.

    Parameters:
        tmp_path: A temporary path to write an empty config to.
    """
    assert main(["--config-file", str(tmp_path / "conf.toml")]) == 0


# IMPORTANT: See top module comment.
def test_show_help(capsys: pytest.CaptureFixture) -> None:
    """Show help.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    with pytest.raises(SystemExit):
        main(["-h"])
    captured = capsys.readouterr()
    assert "git-changelog" in captured.out


def test_get_version() -> None:
    """Get self version."""
    assert get_version()


@pytest.mark.parametrize(
    "args",
    [
        (".", "-s", "feat,fix"),
        ("-s", "feat,fix", "."),
    ],
)
def test_passing_repository_and_sections(tmp_path: Path, args: tuple[str]) -> None:
    """Render the changelog of given repository, choosing sections.

    Parameters:
        tmp_path: A temporary path to write the changelog into.
        args: Command line arguments.
    """
    ch = tmp_path.joinpath("ch.md")
    parsed_settings = parse_settings([*args, "-o", ch.as_posix(), "-c", "angular"])

    assert parsed_settings["output"] == str(ch.as_posix())
    assert parsed_settings["sections"] == ["feat", "fix"]
    assert parsed_settings["repository"] == "."
    assert parsed_settings["convention"] == "angular"


def test_sections_all_parameter(tmp_path: Path) -> None:
    """Test that :all: is passed through correctly from CLI.

    Parameters:
        tmp_path: A temporary path to write the changelog into.
    """
    ch = tmp_path.joinpath("ch.md")
    parsed_settings = parse_settings([".", "-s", ":all:", "-o", ch.as_posix(), "-c", "angular"])

    assert parsed_settings["output"] == str(ch.as_posix())
    assert parsed_settings["sections"] == [":all:"]
    assert parsed_settings["repository"] == "."
    assert parsed_settings["convention"] == "angular"


@pytest.mark.parametrize("is_pyproject", [True, False, None])
@pytest.mark.parametrize(
    ("sections", "sections_value"),
    [
        (None, None),
        ("", None),
        (",,", None),
        ("a, b, ", ["a", "b"]),
        ("a,  , ", ["a"]),
        ("a, b, c", ["a", "b", "c"]),
        (["a", "b", "c"], ["a", "b", "c"]),
        # Uncomment if None/null is once allowed as a value
        # ("none", None),
        # ("none, none, none", None),
    ],
)
@pytest.mark.parametrize("parse_refs", [None, False, True])
def test_config_reading(
    tmp_path: Path,
    is_pyproject: bool | None,
    sections: str | list[str] | None,
    sections_value: list | None,
    parse_refs: bool | None,
) -> None:
    """Check settings files are correctly interpreted.

    Parameters:
        tmp_path: A temporary path to write the settings file into.
        is_pyproject: Controls whether a `pyproject.toml` (`True`),
            a `.git-changelog.toml` (`False`) or a custom file (`None`) is being tested.
        sections: A `sections` config to override defaults.
        sections_value: The expectation for `sections` after reading the config file.
        parse_refs: An explicit override of the `parse_refs` of the config (if boolean)
            or skip writing the override into the test config file (`None`).
    """
    with chdir(str(tmp_path)):
        config_content: dict[str, Any] = {}

        if sections is not None:
            config_content["sections"] = sections

        if parse_refs is not None:
            config_content["parse_refs"] = parse_refs

        config_fname = "custom-file.toml" if is_pyproject is None else ".git-changelog.toml"
        config_fname = "pyproject.toml" if is_pyproject else config_fname
        tmp_path.joinpath(config_fname).write_text(
            tomli_w.dumps(
                config_content if not is_pyproject else {"tool": {"git-changelog": config_content}},
            ),
            encoding="utf8",
        )

        settings = read_config(tmp_path / config_fname) if config_fname == "custom-file.toml" else read_config()

        ground_truth: dict[str, Any] = DEFAULT_SETTINGS.copy()
        ground_truth["sections"] = sections_value
        ground_truth["parse_refs"] = bool(parse_refs)

        assert settings == ground_truth


# YORE: Bump 3: Remove block.
@pytest.mark.parametrize("value", [None, False, True])
def test_settings_warning(
    tmp_path: Path,
    value: bool,
) -> None:
    """Check warning on bump_latest.

    Parameters:
        tmp_path: A temporary path to write the settings file into.
    """
    with chdir(str(tmp_path)):
        args: list[str] = []
        if value is not None:
            (tmp_path / ".git-changelog.toml").write_text(
                tomli_w.dumps({"bump_latest": value}),
                encoding="utf8",
            )
        else:
            args = ["--bump-latest"]

        with pytest.warns(FutureWarning) as record:
            parse_settings(args)

        solution = "is deprecated in favor of"  # Warning comes from CLI parsing.
        if value is not None:  # Warning is issued when parsing the config file.
            solution = "remove" if not value else "auto"

        assert len(record) == 1
        assert solution in str(record[0].message)

        # If setting is in config file AND passed by CLI, two FutureWarnings are issued.
        if (tmp_path / ".git-changelog.toml").exists():
            with pytest.warns(FutureWarning) as record:
                parse_settings(["--bump-latest"])

            assert len(record) == 2


# IMPORTANT: See top module comment.
def test_show_version(capsys: pytest.CaptureFixture) -> None:
    """Show version.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    with pytest.raises(SystemExit):
        main(["-V"])
    captured = capsys.readouterr()
    assert debug._get_version() in captured.out


# IMPORTANT: See top module comment.
def test_show_debug_info(capsys: pytest.CaptureFixture) -> None:
    """Show debug information.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    with pytest.raises(SystemExit):
        main(["--debug-info"])
    captured = capsys.readouterr().out.lower()
    assert "python" in captured
    assert "system" in captured
    assert "environment" in captured
    assert "packages" in captured


# IMPORTANT: See top module comment.
def test_jinja_context(repo: GitRepo) -> None:
    """Render template with custom template variables.

    Parameters:
        repo: Temporary Git repository (fixture).
    """
    repo.path.joinpath("conf.toml").write_text(
        dedent(
            """
            [jinja_context]
            k1 = "ignored"
            k2 = "v2"
            k3 = "v3"
            """,
        ),
        encoding="utf8",
    )

    template = repo.path.joinpath(".custom_template.md.jinja")
    template.write_text(
        "{% for key, val in jinja_context.items() %}{{ key }} = {{ val }}\n{% endfor %}",
        encoding="utf8",
    )

    exit_code = main(
        [
            "--config-file",
            str(repo.path / "conf.toml"),
            "-o",
            str(repo.path / "CHANGELOG.md"),
            "-t",
            f"path:{template}",
            "--jinja-context",
            "k1=v1",
            "-j",
            "k3=v3",
            str(repo.path),
        ],
    )

    assert exit_code == 0

    contents = repo.path.joinpath("CHANGELOG.md").read_text()
    assert contents == "k1 = v1\nk2 = v2\nk3 = v3\n"


# IMPORTANT: See top module comment.
def test_versioning(repo: GitRepo) -> None:
    """Use a specific versioning scheme.

    Parameters:
        repo: Temporary Git repository (fixture).
    """
    repo.commit("feat: Feature")
    repo.tag("1.0.0")
    repo.commit("fix: Fix")
    with repo.enter():
        assert main(["-cconventional", "-nsemver", "-Bauto"]) == 0
        assert main(["-cconventional", "-npep440", "-Bauto"]) == 0
        assert main(["-cconventional", "-nsemver", "-B1.1.0"]) == 0
        assert main(["-cconventional", "-npep440", "-B1.1.0"]) == 0
        assert main(["-cconventional", "-nsemver", "-Bunknown"]) == 1
        assert main(["-cconventional", "-npep440", "-Bunknown"]) == 1
        assert main(["-cconventional", "-npep440", "-Balpha"]) == 1


def test_bumped_version_option(repo: GitRepo, capsys: pytest.CaptureFixture) -> None:
    """The `--bumped-version` CLI option outputs the expected bumped version string."""
    main(
        [
            str(repo.path),
            "--config-file",
            str(repo.path / "conf.toml"),
            "--bump",
            "1.2.3",
            "--bumped-version",
        ],
    )
    captured = capsys.readouterr()
    assert captured.out.strip() == "1.2.3"


def test_include_all_cli_option(tmp_path: Path) -> None:
    """Test that --include-all CLI option is parsed correctly.

    Parameters:
        tmp_path: A temporary path to write an empty config to.
    """
    # Test default (False).
    settings = parse_settings(["--config-file", str(tmp_path / "conf.toml")])
    assert settings["include_all"] is False

    # Test explicit --include-all flag.
    settings = parse_settings(["--config-file", str(tmp_path / "conf.toml"), "--include-all"])
    assert settings["include_all"] is True

    # Test short flag -a.
    settings = parse_settings(["--config-file", str(tmp_path / "conf.toml"), "-a"])
    assert settings["include_all"] is True


def test_include_all_config_option(tmp_path: Path) -> None:
    """Test that include-all config option is read correctly.

    Parameters:
        tmp_path: A temporary path to write config file to.
    """
    with chdir(str(tmp_path)):
        # Test include-all = true in config file.
        (tmp_path / ".git-changelog.toml").write_text(
            tomli_w.dumps({"include-all": True}),
            encoding="utf8",
        )
        settings = read_config()
        assert settings["include_all"] is True

        # Test include-all = false in config file.
        (tmp_path / ".git-changelog.toml").write_text(
            tomli_w.dumps({"include-all": False}),
            encoding="utf8",
        )
        settings = read_config()
        assert settings["include_all"] is False


@pytest.mark.parametrize(
    ("line", "version"),
    [
        ("my-pkg-name (4.2.0~dev) UNRELEASED; urgency=medium", None),
        ("my-pkg-name (4.2.0~dev) stable; urgency=medium", "4.2.0"),
        ("my-pkg-name (4.2.0) stable; urgency=medium", "4.2.0"),
        ("my-pkg-name (4.2.0-1) stable; urgency=medium", "4.2.0"),
        ("my-pkg-name (4.2.0+dfsg-1) stable; urgency=medium", "4.2.0"),
        ("my-pkg-name (4.2.0-alpha1-1) stable; urgency=medium", "4.2.0"),
        ("my-pkg-name (1:4.2.0-1) stable; urgency=medium", "4.2.0"),
        ("my-pkg-name (4.2.0-0progress8.99u1) stable; urgency=medium", "4.2.0"),
    ],
)
def test_debian_version_regex(line: str, version: str) -> None:
    """Test that the Debian version regex matches the expected version string."""
    pattern = re.compile(_DEFAULT_DEBIAN_VERSION_REGEX)
    m = pattern.match(line)
    if version is None:
        assert m is None
    else:
        assert m is not None
        assert m.group("version") == version


def test_outputting_release_notes() -> None:
    """Test that the CLI option to output release notes works as expected."""
    # We make sure that default loaded settings are valid.
    assert main(["--release-notes"]) == 0


@pytest.mark.parametrize("convention", CONVENTIONS)
def test_selecting_convention(convention: str, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that the CLI option to select a convention works as expected."""
    monkeypatch.setattr("git_changelog._internal.cli.build_and_render", lambda *args, **kwargs: None)
    assert main(["-c", convention]) == 0



# Module containing the parsing utilities for git providers.

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from re import Match, Pattern
from typing import ClassVar


class RefRe:
    """An enum helper to store parts of regular expressions for references."""

    BB = r"(?:^|[\s,])"  # blank before
    """Blank before the reference."""
    BA = r"(?:[\s,]|$)"  # blank after
    """Blank after the reference."""
    NP = r"(?:(?P<namespace>[-\w]+)/)?(?P<project>[-\w]+)"  # namespace and project
    """Namespace and project."""
    ID = r"{symbol}(?P<ref>[1-9]\d*)"
    """Issue or pull request ID."""
    ONE_WORD = r"{symbol}(?P<ref>\w*[-a-z_ ][-\w]*)"
    """One word reference."""
    MULTI_WORD = r'{symbol}(?P<ref>"\w[- \w]*")'
    """Multi word reference."""
    COMMIT = r"(?P<ref>[0-9a-f]{{{min},{max}}})"
    """Commit hash."""
    COMMIT_RANGE = r"(?P<ref>[0-9a-f]{{{min},{max}}}\.\.\.[0-9a-f]{{{min},{max}}})"
    """Commit range."""
    MENTION = r"@(?P<ref>\w[-\w]*)"
    """Mention."""


class Ref:
    """A class to represent a reference and its URL."""

    def __init__(self, ref: str, url: str) -> None:
        """Initialization method.

        Parameters:
            ref: The reference text.
            url: The reference URL.
        """
        self.ref: str = ref
        """The reference text."""
        self.url: str = url
        """The reference URL."""

    def __str__(self):
        return self.ref + ": " + self.url


class RefDef:
    """A class to store a reference regular expression and URL building string."""

    def __init__(self, regex: Pattern, url_string: str):
        """Initialization method.

        Parameters:
            regex: The regular expression to match the reference.
            url_string: The URL string to format using matched groups.
        """
        self.regex = regex
        """The regular expression to match the reference."""
        self.url_string = url_string
        """The URL string to format using matched groups."""


class ProviderRefParser(ABC):
    """A base class for specific providers reference parsers."""

    url: str
    """The base URL for the provider."""
    namespace: str
    """The namespace for the provider."""
    project: str
    """The project for the provider."""
    REF: ClassVar[dict[str, RefDef]] = {}
    """The reference definitions for the provider."""

    def __init__(self, namespace: str, project: str, url: str | None = None):
        """Initialization method.

        Parameters:
            namespace: The Bitbucket namespace.
            project: The Bitbucket project.
            url: The Bitbucket URL.
        """
        self.namespace: str = namespace
        """The namespace for the provider."""
        self.project: str = project
        """The project for the provider."""
        self.url: str = url or self.url
        """The URL for the provider."""

    def get_refs(self, ref_type: str, text: str) -> list[Ref]:
        """Find all references in the given text.

        Parameters:
            ref_type: The reference type.
            text: The text in which to search references.

        Returns:
            A list of references (instances of [Ref][git_changelog.Ref]).
        """
        return [
            Ref(ref=match.group().strip(), url=self.build_ref_url(ref_type, match.groupdict()))
            for match in self.parse_refs(ref_type, text)
        ]

    def parse_refs(self, ref_type: str, text: str) -> list[Match]:
        """Parse references in the given text.

        Parameters:
            ref_type: The reference type.
            text: The text to parse.

        Returns:
            A list of regular expressions matches.
        """
        if ref_type not in self.REF:
            refs = [key for key in self.REF if key.startswith(ref_type)]
            return [match for ref in refs for match in self.REF[ref].regex.finditer(text)]
        return list(self.REF[ref_type].regex.finditer(text))

    def build_ref_url(self, ref_type: str, match_dict: dict[str, str]) -> str:
        """Build the URL for a reference type and a dictionary of matched groups.

        Parameters:
            ref_type: The reference type.
            match_dict: The matched groups.

        Returns:
            The built URL.
        """
        return self.REF[ref_type].url_string.format(**match_dict)

    @abstractmethod
    def get_tag_url(self, tag: str) -> str:
        """Get the URL for a git tag.

        Parameters:
            tag: The git tag.

        Returns:
            The tag URL.
        """
        raise NotImplementedError

    @abstractmethod
    def get_compare_url(self, base: str, target: str) -> str:
        """Get the URL for a tag comparison.

        Parameters:
            base: The base tag.
            target: The target tag.

        Returns:
            The comparison URL.
        """
        raise NotImplementedError


class GitHub(ProviderRefParser):
    """A parser for the GitHub references."""

    url: str = "https://github.com"
    """The base URL for the provider."""
    project_url: str = "{base_url}/{namespace}/{project}"
    """The project URL for the provider."""
    tag_url: str = "{base_url}/{namespace}/{project}/releases/tag/{ref}"
    """The tag URL for the provider."""

    commit_min_length = 8
    """The minimum length of a commit hash."""
    commit_max_length = 40
    """The maximum length of a commit hash."""

    REF: ClassVar[dict[str, RefDef]] = {
        "issues": RefDef(
            regex=re.compile(RefRe.BB + RefRe.NP + "?" + RefRe.ID.format(symbol="#"), re.IGNORECASE),
            url_string="{base_url}/{namespace}/{project}/issues/{ref}",
        ),
        "commits": RefDef(
            regex=re.compile(
                RefRe.BB
                + r"(?:{np}@)?{commit}{ba}".format(
                    np=RefRe.NP,
                    commit=RefRe.COMMIT.format(min=commit_min_length, max=commit_max_length),
                    ba=RefRe.BA,
                ),
                re.IGNORECASE,
            ),
            url_string="{base_url}/{namespace}/{project}/commit/{ref}",
        ),
        "commits_ranges": RefDef(
            regex=re.compile(
                RefRe.BB
                + r"(?:{np}@)?{commit_range}".format(
                    np=RefRe.NP,
                    commit_range=RefRe.COMMIT_RANGE.format(min=commit_min_length, max=commit_max_length),
                ),
                re.IGNORECASE,
            ),
            url_string="{base_url}/{namespace}/{project}/compare/{ref}",
        ),
        "mentions": RefDef(regex=re.compile(RefRe.BB + RefRe.MENTION, re.IGNORECASE), url_string="{base_url}/{ref}"),
    }
    """The reference definitions for the provider."""

    def build_ref_url(self, ref_type: str, match_dict: dict[str, str]) -> str:
        match_dict["base_url"] = self.url
        if not match_dict.get("namespace"):
            match_dict["namespace"] = self.namespace
        if not match_dict.get("project"):
            match_dict["project"] = self.project
        return super().build_ref_url(ref_type, match_dict)

    def get_tag_url(self, tag: str = "") -> str:
        return self.tag_url.format(base_url=self.url, namespace=self.namespace, project=self.project, ref=tag)

    def get_compare_url(self, base: str, target: str) -> str:
        return self.build_ref_url("commits_ranges", {"ref": f"{base}...{target}"})


class GitLab(ProviderRefParser):
    """A parser for the GitLab references."""

    url: str = "https://gitlab.com"
    """The base URL for the provider."""
    project_url: str = "{base_url}/{namespace}/{project}"
    """The project URL for the provider."""
    tag_url: str = "{base_url}/{namespace}/{project}/tags/{ref}"
    """The tag URL for the provider."""

    commit_min_length = 8
    """The minimum length of a commit hash."""
    commit_max_length = 40
    """The maximum length of a commit hash."""

    REF: ClassVar[dict[str, RefDef]] = {
        "issues": RefDef(
            regex=re.compile(RefRe.BB + RefRe.NP + "?" + RefRe.ID.format(symbol="#"), re.IGNORECASE),
            url_string="{base_url}/{namespace}/{project}/issues/{ref}",
        ),
        "merge_requests": RefDef(
            regex=re.compile(RefRe.BB + RefRe.NP + "?" + RefRe.ID.format(symbol=r"!"), re.IGNORECASE),
            url_string="{base_url}/{namespace}/{project}/merge_requests/{ref}",
        ),
        "snippets": RefDef(
            regex=re.compile(RefRe.BB + RefRe.NP + "?" + RefRe.ID.format(symbol=r"\$"), re.IGNORECASE),
            url_string="{base_url}/{namespace}/{project}/snippets/{ref}",
        ),
        "labels_ids": RefDef(
            regex=re.compile(RefRe.BB + RefRe.NP + "?" + RefRe.ID.format(symbol=r"~"), re.IGNORECASE),
            url_string="{base_url}/{namespace}/{project}/issues?label_name[]={ref}",  # no label_id param?
        ),
        "labels_one_word": RefDef(
            regex=re.compile(  # also matches label IDs
                RefRe.BB + RefRe.NP + "?" + RefRe.ONE_WORD.format(symbol=r"~"),
                re.IGNORECASE,
            ),
            url_string="{base_url}/{namespace}/{project}/issues?label_name[]={ref}",
        ),
        "labels_multi_word": RefDef(
            regex=re.compile(RefRe.BB + RefRe.NP + "?" + RefRe.MULTI_WORD.format(symbol=r"~"), re.IGNORECASE),
            url_string="{base_url}/{namespace}/{project}/issues?label_name[]={ref}",
        ),
        "milestones_ids": RefDef(
            regex=re.compile(  # also matches milestones IDs
                RefRe.BB + RefRe.NP + "?" + RefRe.ID.format(symbol=r"%"),
                re.IGNORECASE,
            ),
            url_string="{base_url}/{namespace}/{project}/milestones/{ref}",
        ),
        "milestones_one_word": RefDef(
            regex=re.compile(RefRe.BB + RefRe.NP + "?" + RefRe.ONE_WORD.format(symbol=r"%"), re.IGNORECASE),
            url_string="{base_url}/{namespace}/{project}/milestones",  # cannot guess ID
        ),
        "milestones_multi_word": RefDef(
            regex=re.compile(RefRe.BB + RefRe.NP + "?" + RefRe.MULTI_WORD.format(symbol=r"%"), re.IGNORECASE),
            url_string="{base_url}/{namespace}/{project}/milestones",  # cannot guess ID
        ),
        "commits": RefDef(
            regex=re.compile(
                RefRe.BB
                + r"(?:{np}@)?{commit}{ba}".format(
                    np=RefRe.NP,
                    commit=RefRe.COMMIT.format(min=commit_min_length, max=commit_max_length),
                    ba=RefRe.BA,
                ),
                re.IGNORECASE,
            ),
            url_string="{base_url}/{namespace}/{project}/commit/{ref}",
        ),
        "commits_ranges": RefDef(
            regex=re.compile(
                RefRe.BB
                + r"(?:{np}@)?{commit_range}".format(
                    np=RefRe.NP,
                    commit_range=RefRe.COMMIT_RANGE.format(min=commit_min_length, max=commit_max_length),
                ),
                re.IGNORECASE,
            ),
            url_string="{base_url}/{namespace}/{project}/compare/{ref}",
        ),
        "mentions": RefDef(regex=re.compile(RefRe.BB + RefRe.MENTION, re.IGNORECASE), url_string="{base_url}/{ref}"),
    }
    """The reference definitions for the provider."""

    def build_ref_url(self, ref_type: str, match_dict: dict[str, str]) -> str:
        match_dict["base_url"] = self.url
        if not match_dict.get("namespace"):
            match_dict["namespace"] = self.namespace
        if not match_dict.get("project"):
            match_dict["project"] = self.project
        if ref_type.startswith("label"):
            match_dict["ref"] = match_dict["ref"].replace('"', "").replace(" ", "+")
        return super().build_ref_url(ref_type, match_dict)

    def get_tag_url(self, tag: str = "") -> str:
        return self.tag_url.format(base_url=self.url, namespace=self.namespace, project=self.project, ref=tag)

    def get_compare_url(self, base: str, target: str) -> str:
        return self.build_ref_url("commits_ranges", {"ref": f"{base}...{target}"})


class Bitbucket(ProviderRefParser):
    """A parser for the Bitbucket references."""

    url: str = "https://bitbucket.org"
    """The base URL for the provider."""
    project_url: str = "{base_url}/{namespace}/{project}"
    """The project URL for the provider."""
    tag_url: str = "{base_url}/{namespace}/{project}/commits/tag/{ref}"
    """The tag URL for the provider."""

    commit_min_length = 8
    """The minimum length of a commit hash."""
    commit_max_length = 40
    """The maximum length of a commit hash."""

    REF: ClassVar[dict[str, RefDef]] = {
        "issues": RefDef(
            regex=re.compile(RefRe.BB + RefRe.NP + "?issue\\s*" + RefRe.ID.format(symbol="#"), re.IGNORECASE),
            url_string="{base_url}/{namespace}/{project}/issues/{ref}",
        ),
        "merge_requests": RefDef(
            regex=re.compile(RefRe.BB + RefRe.NP + "?pull request\\s*" + RefRe.ID.format(symbol=r"#"), re.IGNORECASE),
            url_string="{base_url}/{namespace}/{project}/pull-request/{ref}",
        ),
        "commits": RefDef(
            regex=re.compile(
                RefRe.BB
                + r"(?:{np}@)?{commit}{ba}".format(
                    np=RefRe.NP,
                    commit=RefRe.COMMIT.format(min=commit_min_length, max=commit_max_length),
                    ba=RefRe.BA,
                ),
                re.IGNORECASE,
            ),
            url_string="{base_url}/{namespace}/{project}/commits/{ref}",
        ),
        "commits_ranges": RefDef(
            regex=re.compile(
                RefRe.BB
                + r"(?:{np}@)?{commit_range}".format(
                    np=RefRe.NP,
                    commit_range=RefRe.COMMIT_RANGE.format(min=commit_min_length, max=commit_max_length),
                ),
                re.IGNORECASE,
            ),
            url_string="{base_url}/{namespace}/{project}/branches/compare/{ref}#diff",
        ),
        "mentions": RefDef(
            regex=re.compile(RefRe.BB + RefRe.MENTION, re.IGNORECASE),
            url_string="{base_url}/{ref}",
        ),
    }
    """The reference definitions for the provider."""

    def build_ref_url(self, ref_type: str, match_dict: dict[str, str]) -> str:
        match_dict["base_url"] = self.url
        if not match_dict.get("namespace"):
            match_dict["namespace"] = self.namespace
        if not match_dict.get("project"):
            match_dict["project"] = self.project
        return super().build_ref_url(ref_type, match_dict)

    def get_tag_url(self, tag: str = "") -> str:
        return self.tag_url.format(base_url=self.url, namespace=self.namespace, project=self.project, ref=tag)

    def get_compare_url(self, base: str, target: str) -> str:
        return self.build_ref_url("commits_ranges", {"ref": f"{target}..{base}"})



# The subpackage containing the builtin templates.

from __future__ import annotations

from pathlib import Path
from urllib.parse import urlparse

from jinja2 import Environment, Template

TEMPLATES_PATH = Path(__file__).parent
"""The path to the templates directory."""
JINJA_ENV = Environment()  # noqa: S701
"""The Jinja environment."""


def _filter_is_url(value: str) -> bool:
    return bool(urlparse(value).scheme)


def configure_env(env: Environment) -> None:
    """Configure the Jinja environment.

    Parameters:
        env: The environment to configure.
    """
    env.filters.update({"is_url": _filter_is_url})


def get_custom_template(path: str | Path) -> Template:
    """Get a custom template instance.

    Parameters:
        path: Path to the custom template.

    Returns:
        The Jinja template.
    """
    return JINJA_ENV.from_string(Path(path).read_text(encoding="utf8"))


def get_template(name: str) -> Template:
    """Get a builtin template instance.

    Parameters:
        name: The template name.

    Returns:
        The Jinja template.
    """
    return JINJA_ENV.from_string(TEMPLATES_PATH.joinpath(f"{name}.md.jinja").read_text(encoding="utf8"))


configure_env(JINJA_ENV)



# Utilities to handle different versioning schemes such as SemVer and PEP 440.

from __future__ import annotations

import sys
from typing import Any, Literal, Protocol

import semver
from packaging import version as packaging_version

# YORE: EOL 3.13: Replace block with line 2.
if sys.version_info >= (3, 13):
    from copy import replace
else:

    def replace(obj: Any, /, **changes: Any) -> Any:
        cls = obj.__class__
        func = getattr(cls, "__replace__", None)
        if func is None:
            raise TypeError(f"replace() does not support {cls.__name__} objects")
        return func(obj, **changes)


SemVerStrategy = Literal[
    "major",
    "minor",
    "patch",
    "release",
]
"""SemVer versioning strategies."""
PEP440Strategy = Literal[
    "epoch",
    "release",
    "major",
    "minor",
    "micro",
    "patch",
    "pre",
    "alpha",
    "beta",
    "candidate",
    "post",
    "dev",
    "major+alpha",
    "major+beta",
    "major+candidate",
    "major+dev",
    "major+alpha+dev",
    "major+beta+dev",
    "major+candidate+dev",
    "minor+alpha",
    "minor+beta",
    "minor+candidate",
    "minor+dev",
    "minor+alpha+dev",
    "minor+beta+dev",
    "minor+candidate+dev",
    "micro+alpha",
    "micro+beta",
    "micro+candidate",
    "micro+dev",
    "micro+alpha+dev",
    "micro+beta+dev",
    "micro+candidate+dev",
    "alpha+dev",
    "beta+dev",
    "candidate+dev",
]
"""PEP 440 versioning strategies."""


_release_kind = {"a": "alpha", "b": "beta", "c": "candidate", "rc": "candidate", "p": "post"}


class ParsedVersion(Protocol):  # noqa: PLW1641
    """Base class for versioning schemes."""

    def __lt__(self, other: object) -> bool:
        """Implement `<` comparison."""
        ...

    def __le__(self, other: object) -> bool:
        """Implement `<=` comparison."""
        ...

    def __eq__(self, other: object) -> bool:
        """Implement `==` comparison."""
        ...

    def __ge__(self, other: object) -> bool:
        """Implement `>=` comparison."""
        ...

    def __gt__(self, other: object) -> bool:
        """Implement `>` comparison."""
        ...

    def __ne__(self, other: object) -> bool:
        """Implement `!=` comparison."""
        ...


class SemVerVersion(semver.Version, ParsedVersion):
    """SemVer version."""

    def bump_major(self) -> SemVerVersion:
        """Bump the major version."""
        return SemVerVersion(*super().bump_major())  # ty:ignore[not-iterable]

    def bump_minor(self) -> SemVerVersion:
        """Bump the minor version."""
        return SemVerVersion(*super().bump_minor())  # ty:ignore[not-iterable]

    def bump_patch(self) -> SemVerVersion:
        """Bump the patch version."""
        return SemVerVersion(*super().bump_patch())  # ty:ignore[not-iterable]

    def bump_release(self) -> SemVerVersion:
        """Bump from a pre-release to the same, regular release.

        Returns:
            The same version, without pre-release or build metadata.
        """
        return SemVerVersion(self.major, self.minor, self.patch)


class PEP440Version(packaging_version.Version, ParsedVersion):
    """PEP 440 version."""

    @classmethod
    def from_parts(
        cls,
        epoch: int | None = None,
        release: tuple[int, ...] | None = None,
        pre: tuple[Literal["a", "b", "rc"], int] | None = None,
        post: int | None = None,
        dev: int | None = None,
    ) -> PEP440Version:
        """Build a version from its parts.

        Parameters:
            epoch: Version's epoch number.
            release: Version's release numbers.
            pre: Version's prerelease kind and number.
            post: Version's post number.
            dev: Version's dev number.

        Returns:
            A PEP 440 version.
        """
        # Since the original class only allows instantiating a version
        # by passing a string, we first create a dummy version "1"
        # and then use __replace__ to set the real version parts.
        return replace(
            cls("1"),
            epoch=epoch or 0,
            release=release or (),
            pre=pre,
            post=post,
            dev=dev,
            local=None,
        )

    def bump_epoch(self) -> PEP440Version:
        """Bump epoch.

        Examples:
            >>> PEP440Version("1.0").bump_epoch()
            <Version('1!1.0')>
            >>> PEP440Version("0!1.0").bump_epoch()
            <Version('1!1.0')>
            >>> PEP440Version("1!1.0").bump_epoch()
            <Version('2!1.0')>
            >>> PEP440Version("1.0a2.post3").bump_epoch()
            <Version('2!1.0')>

        Returns:
            Version with bumped epoch, same release, and the right parts reset to 0 or nothing.
        """
        return PEP440Version.from_parts(epoch=self.epoch + 1, release=self.release)

    def bump_release(self, level: int | None = None, *, trim: bool = False) -> PEP440Version:
        """Bump given release level.

        Parameters:
            level: The release level to bump.

                - 0 means major
                - 1 means minor
                - 2 means micro (patch)
                - 3+ don't have names
                - None means move from pre-release to release (unchanged)
            trim: Whether to trim all zeroes on the right after bumping.

        Examples:
            >>> PEP440Version("1").bump_release(0)
            <Version('2')>
            >>> PEP440Version("1.1").bump_release(0)
            <Version('2.0')>
            >>> PEP440Version("1.1.1").bump_release(0)
            <Version('2.0.0')>
            >>> PEP440Version("1.1.1").bump_release(0, trim=True)
            <Version('2')>
            >>> PEP440Version("1a2.post3").bump_release(0)
            <Version('2')>

            >>> PEP440Version("1").bump_release(1)
            <Version('1.1')>
            >>> PEP440Version("1.1").bump_release(1)
            <Version('1.2')>
            >>> PEP440Version("1.1.1").bump_release(1)
            <Version('1.2.0')>
            >>> PEP440Version("1.1.1").bump_release(1, trim=True)
            <Version('1.2')>
            >>> PEP440Version("1a2.post3").bump_release(1)
            <Version('1.1')>

            >>> PEP440Version("1a0").bump_release()
            <Version('1')>
            >>> PEP440Version("1b1").bump_release()
            <Version('1')>
            >>> PEP440Version("1rc2").bump_release()
            <Version('1')>
            >>> PEP440Version("1a2.dev0").bump_release()
            <Version('1')>
            >>> PEP440Version("1post0").bump_release()
            ValueError: Cannot bump from post-release to release

        Returns:
            Version with same epoch, bumped release level, and the right parts reset to 0 or nothing.
        """
        release = list(self.release)

        # When level is not specified, user wants to bump the version
        # as a "release", going out of alpha/beta/candidate phase.
        # So we simply keep the release part as it is, optionally trimming it.
        if level is None:
            # However if this is a post-release, this is an error:
            # we can't bump from a post-release to the same, regular release.
            if self.post is not None:
                raise ValueError("Cannot bump from post-release to release")
            if trim:
                while release[-1] == 0:
                    release.pop()

        # When level is specified, we bump the specified level.
        # If the given level is higher that the number of release parts,
        # we insert the missing parts as 0.
        else:
            try:
                release[level] += 1
            except IndexError:
                while len(release) < level:
                    release.append(0)
                release.append(1)
            if trim:
                release = release[: level + 1]
            else:
                for index in range(level + 1, len(release)):
                    release[index] = 0

        # We rebuild the version with same epoch, updated release,
        # and pre/post/dev parts dropped.
        return PEP440Version.from_parts(epoch=self.epoch, release=tuple(release))

    def bump_major(self, *, trim: bool = False) -> PEP440Version:
        """Bump major.

        Parameters:
            trim: Whether to trim all zeroes on the right after bumping.

        Examples:
            >>> PEP440Version("1").bump_major()
            <Version('2')>
            >>> PEP440Version("1.1").bump_major()
            <Version('2.0')>
            >>> PEP440Version("1.1.1").bump_major()
            <Version('2.0.0')>
            >>> PEP440Version("1.1.1").bump_major(trim=True)
            <Version('2')>
            >>> PEP440Version("1a2.post3").bump_major()
            <Version('2')>

        Returns:
            Version with same epoch, bumped major and the right parts reset to 0 or nothing.
        """
        return self.bump_release(level=0, trim=trim)

    def bump_minor(self, *, trim: bool = False) -> PEP440Version:
        """Bump minor.

        Parameters:
            trim: Whether to trim all zeroes on the right after bumping.

        Examples:
            >>> PEP440Version("1").bump_minor()
            <Version('1.1')>
            >>> PEP440Version("1.1").bump_minor()
            <Version('1.2')>
            >>> PEP440Version("1.1.1").bump_minor()
            <Version('1.2.0')>
            >>> PEP440Version("1.1.1").bump_minor(trim=True)
            <Version('1.2')>
            >>> PEP440Version("1a2.post3").bump_minor()
            <Version('1.1')>

        Returns:
            Version with same epoch, same major, bumped minor and the right parts reset to 0 or nothing.
        """
        return self.bump_release(level=1, trim=trim)

    def bump_micro(self, *, trim: bool = False) -> PEP440Version:
        """Bump micro.

        Parameters:
            trim: Whether to trim all zeroes on the right after bumping.

        Examples:
            >>> PEP440Version("1").bump_micro()
            <Version('1.0.1')>
            >>> PEP440Version("1.1").bump_micro()
            <Version('1.1.1')>
            >>> PEP440Version("1.1.1").bump_micro()
            <Version('1.1.2')>
            >>> PEP440Version("1.1.1").bump_micro()
            <Version('1.1.2')>
            >>> PEP440Version("1.1.1.1").bump_micro()
            <Version('1.1.2.0')>
            >>> PEP440Version("1.1.1.1").bump_micro(trim=True)
            <Version('1.1.2')>
            >>> PEP440Version("1a2.post3").bump_micro()
            <Version('1.0.1')>

        Returns:
            Version with same epoch, same major, same minor, bumped micro and the right parts reset to 0 or nothing.
        """
        return self.bump_release(level=2, trim=trim)

    def bump_pre(self, pre: Literal["a", "b", "rc"] | None = None) -> PEP440Version:
        """Bump pre-release.

        Parameters:
            pre: Kind of pre-release to bump.

                - a means alpha
                - b means beta
                - c or rc means (release) candidate

        Examples:
            >>> PEP440Version("1").bump_pre()
            ValueError: Cannot bump from release to alpha pre-release (use `dent_pre`)
            >>> PEP440Version("1a0").bump_pre()
            <Version('1a1')>
            >>> PEP440Version("1a0").bump_pre("a")
            <Version('1a1')>
            >>> PEP440Version("1a0.post0").bump_pre()
            <Version('1a1')>
            >>> PEP440Version("1b2").bump_pre("a")
            ValueError: Cannot bump from beta to alpha pre-release (use `dent_alpha`)
            >>> PEP440Version("1c2").bump_pre("a")
            ValueError: Cannot bump from candidate to alpha pre-release (use `dent_alpha`)

            >>> PEP440Version("1").bump_pre("b")
            ValueError: Cannot bump from release to beta pre-release (use `dent_beta`)
            >>> PEP440Version("1a2").bump_pre("b")
            <Version('1b0')>
            >>> PEP440Version("1b2").bump_pre("b")
            <Version('1b3')>
            >>> PEP440Version("1c2").bump_pre("b")
            ValueError: Cannot bump from candidate to beta pre-release (use `dent_beta`)

            >>> PEP440Version("1").bump_pre("c")
            ValueError: Cannot bump from release to candidate pre-release (use `dent_candidate`)
            >>> PEP440Version("1a2").bump_pre("c")
            <Version('1rc0')>
            >>> PEP440Version("1b2").bump_pre("rc")
            <Version('1rc0')>
            >>> PEP440Version("1rc2").bump_pre("c")
            <Version('1rc3')>

        Returns:
            Version with same epoch, same release, bumped pre-release and the right parts reset to 0 or nothing.
        """
        if self.pre is None:
            kind = _release_kind.get(pre, "")  # ty:ignore[no-matching-overload]
            raise ValueError(
                f"Cannot bump from release to {kind + ' ' if kind else ''}pre-release (use `dent_{kind or 'pre'}`)",
            )
        current_pre: Literal["a", "b", "rc"]
        current_pre, number = self.pre  # ty:ignore[invalid-assignment]
        if pre is None:
            pre = current_pre
        if pre == current_pre:
            number += 1
        elif current_pre < pre:
            number = 0
        else:
            raise ValueError(
                f"Cannot bump from {_release_kind.get(current_pre, 'release')} to {_release_kind[pre]} pre-release (use `dent_{_release_kind[pre]}`)",
            )
        return PEP440Version.from_parts(epoch=self.epoch, release=self.release, pre=(pre, number))

    def bump_alpha(self) -> PEP440Version:
        """Bump alpha-release.

        Examples:
            >>> PEP440Version("1").bump_alpha()
            ValueError: Cannot bump from release to alpha pre-release (use `dent_alpha`)
            >>> PEP440Version("1a0").bump_alpha()
            <Version('1a1')>
            >>> PEP440Version("1a0").bump_alpha("a")
            <Version('1a1')>
            >>> PEP440Version("1a0.post0").bump_alpha()
            <Version('1a1')>

        Returns:
            Version with same epoch, same release, bumped alpha pre-release and the right parts reset to 0 or nothing.
        """
        return self.bump_pre("a")

    def bump_beta(self) -> PEP440Version:
        """Bump beta-release.

        Examples:
            >>> PEP440Version("1").bump_beta()
            ValueError: Cannot bump from release to beta pre-release (use `dent_beta`)
            >>> PEP440Version("1b0").bump_beta()
            <Version('1b1')>
            >>> PEP440Version("1b0").bump_beta()
            <Version('1b1')>
            >>> PEP440Version("1b0.post0").bump_beta()
            <Version('1b1')>

        Returns:
            Version with same epoch, same release, bumped beta pre-release and the right parts reset to 0 or nothing.
        """
        return self.bump_pre("b")

    def bump_candidate(self) -> PEP440Version:
        """Bump candidate release.

        Examples:
            >>> PEP440Version("1").bump_candidate()
            ValueError: Cannot bump from release to candidate pre-release (use `dent_candidate`)
            >>> PEP440Version("1c0").bump_candidate()
            <Version('1rc1')>
            >>> PEP440Version("1c0").bump_candidate()
            <Version('1rc1')>
            >>> PEP440Version("1c0.post0").bump_candidate()
            <Version('1rc1')>

        Returns:
            Version with same epoch, same release, bumped candidate pre-release and the right parts reset to 0 or nothing.
        """
        return self.bump_pre("rc")

    def bump_post(self) -> PEP440Version:
        """Bump post-release.

        Examples:
            >>> PEP440Version("1").bump_post()
            <Version('1.post0')>
            >>> PEP440Version("1.post0").bump_post()
            <Version('1.post1')>
            >>> PEP440Version("1a0.post0").bump_post()
            <Version('1a0.post1')>
            >>> PEP440Version("1.post0.dev1").bump_post()
            <Version('1.post1')>

        Returns:
            Version with same epoch, same release, same pre-release, bumped post-release and the right parts reset to 0 or nothing.
        """
        post = 0 if self.post is None else self.post + 1
        return PEP440Version.from_parts(
            epoch=self.epoch,
            release=self.release,
            pre=self.pre,  # ty:ignore[invalid-argument-type]
            post=post,
        )

    def bump_dev(self) -> PEP440Version:
        """Bump dev-release.

        Examples:
            >>> PEP440Version("1").bump_dev()
            ValueError: Cannot bump from release to dev-release (use `dent_dev`)
            >>> PEP440Version("1a0").bump_dev()
            ValueError: Cannot bump from alpha to dev-release (use `dent_dev`)
            >>> PEP440Version("1b1").bump_dev()
            ValueError: Cannot bump from beta to dev-release (use `dent_dev`)
            >>> PEP440Version("1rc2").bump_dev()
            ValueError: Cannot bump from candidate to dev-release (use `dent_dev`)
            >>> PEP440Version("1.post0").bump_dev()
            ValueError: Cannot bump from post to dev-release (use `dent_dev`)
            >>> PEP440Version("1a0.dev1").bump_dev()
            <Version('1a0.dev2')>

        Returns:
            Version with same epoch, same release, same pre-release, same post-release and bumped dev-release.
        """
        if self.dev is None:
            if self.post is not None:
                kind = "p"
            elif self.pre is not None:
                kind = self.pre[0]
            else:
                kind = "z"
            raise ValueError(f"Cannot bump from {_release_kind.get(kind, 'release')} to dev-release (use `dent_dev`)")
        return PEP440Version.from_parts(
            epoch=self.epoch,
            release=self.release,
            pre=self.pre,  # ty:ignore[invalid-argument-type]
            post=self.post,
            dev=self.dev + 1,
        )

    def dent_pre(self, pre: Literal["a", "b", "rc"] | None = None) -> PEP440Version:
        """Dent to pre-release.

        This method dents a release down to an alpha, beta or candidate pre-release.

        Parameters:
            pre: Kind of pre-release to bump.

                - a means alpha
                - b means beta
                - c or rc means (release) candidate

        Examples:
            >>> PEP440Version("1").dent_pre()
            <Version('1a0')>
            >>> PEP440Version("1").dent_pre("a")
            <Version('1a0')>
            >>> PEP440Version("1a0").dent_pre("a")
            ValueError: Cannot dent alpha pre-releases
            >>> PEP440Version("1").dent_pre("b")
            <Version('1b0')>
            >>> PEP440Version("1b0").dent_pre("b")
            ValueError: Cannot dent beta pre-releases
            >>> PEP440Version("1").dent_pre("c")
            <Version('1rc0')>
            >>> PEP440Version("1").dent_pre("rc")
            <Version('1rc0')>
            >>> PEP440Version("1rc0").dent_pre("c")
            ValueError: Cannot dent candidate pre-releases

        Returns:
            Version with same epoch and release dented to pre-release.
        """
        if self.pre is not None:
            raise ValueError(f"Cannot dent {_release_kind[self.pre[0]]} pre-releases")
        if pre is None:
            pre = "a"
        return PEP440Version.from_parts(epoch=self.epoch, release=self.release, pre=(pre, 0))

    def dent_alpha(self) -> PEP440Version:
        """Dent to alpha-release.

        Examples:
            >>> PEP440Version("1").dent_alpha()
            <Version('1a0')>
            >>> PEP440Version("1a0").dent_alpha()
            ValueError: Cannot dent alpha pre-releases
            >>> PEP440Version("1b0").dent_alpha()
            ValueError: Cannot dent beta pre-releases
            >>> PEP440Version("1rc0").dent_alpha()
            ValueError: Cannot dent candidate pre-releases

        Returns:
            Version with same epoch and release dented to alpha pre-release.
        """
        return self.dent_pre("a")

    def dent_beta(self) -> PEP440Version:
        """Dent to beta-release.

        Examples:
            >>> PEP440Version("1").dent_beta()
            <Version('1b0')>
            >>> PEP440Version("1a0").dent_beta()
            ValueError: Cannot dent alpha pre-releases
            >>> PEP440Version("1b0").dent_beta()
            ValueError: Cannot dent beta pre-releases
            >>> PEP440Version("1rc0").dent_beta()
            ValueError: Cannot dent candidate pre-releases

        Returns:
            Version with same epoch and release dented to beta pre-release.
        """
        return self.dent_pre("b")

    def dent_candidate(self) -> PEP440Version:
        """Dent to candidate release.

        Examples:
            >>> PEP440Version("1").dent_candidate()
            <Version('1rc0')>
            >>> PEP440Version("1a0").dent_candidate()
            ValueError: Cannot dent alpha pre-releases
            >>> PEP440Version("1b0").dent_candidate()
            ValueError: Cannot dent beta pre-releases
            >>> PEP440Version("1rc0").dent_candidate()
            ValueError: Cannot dent candidate pre-releases

        Returns:
            Version with same epoch and release dented to candidate pre-release.
        """
        return self.dent_pre("rc")

    def dent_dev(self) -> PEP440Version:
        """Dent to dev-release.

        Examples:
            >>> PEP440Version("1").dent_dev()
            <Version('1.dev0')>
            >>> PEP440Version("1a0").dent_dev()
            <Version('1a0.dev0')>
            >>> PEP440Version("1b1").dent_dev()
            <Version('1b1.dev0')>
            >>> PEP440Version("1c2").dent_dev()
            <Version('1rc2.dev0')>
            >>> PEP440Version("1.post0").dent_dev()
            <Version('1.post0.dev0')>
            >>> PEP440Version("1a0.dev1").dent_dev()
            ValueError: Cannot dent dev-releases

        Returns:
            Version with same epoch and release dented to dev-release.
        """
        if self.dev is not None:
            raise ValueError("Cannot dent dev-releases")
        return PEP440Version.from_parts(
            epoch=self.epoch,
            release=self.release,
            pre=self.pre,  # ty:ignore[invalid-argument-type]
            post=self.post,
            dev=0,
        )


def version_prefix(version: str) -> tuple[str, str]:
    """Return a version and its optional `v` prefix.

    Parameters:
        version: The full version.

    Returns:
        version: The version without its prefix.
        prefix: The version prefix.
    """
    prefix = ""
    if version[0] == "v":
        prefix = "v"
        version = version[1:]
    return version, prefix


def parse_semver(version: str) -> tuple[SemVerVersion, str]:
    """Parse a SemVer version.

    Returns:
        A semver version instance with useful methods.
    """
    version, prefix = version_prefix(version)
    return SemVerVersion.parse(version), prefix


def parse_pep440(version: str) -> tuple[PEP440Version, str]:
    """Parse a PEP version.

    Returns:
        A PEP 440 version instance with useful methods.
    """
    version, prefix = version_prefix(version)
    return PEP440Version(version), prefix


class VersionBumper:
    """Base class for version bumpers."""

    initial: str
    """The initial version."""

    def __init__(self, strategies: tuple[str, ...]) -> None:
        """Initialize the bumper.

        Parameters:
            strategies: The supported bumping strategies.
        """
        self.strategies = strategies
        """The supported bumping strategies."""

    def __call__(self, version: str, strategy: str, **kwargs: Any) -> str:
        """Bump a version.

        Parameters:
            version: The version to bump.
            strategy: The bumping strategy.
            **kwargs: Additional bumper-specific arguments.

        Returns:
            The bumped version.
        """
        raise NotImplementedError


class PEP440Bumper(VersionBumper):
    """PEP 440 version bumper."""

    initial: str = "0.0.0"
    """The initial version."""

    def __call__(  # ty:ignore[invalid-method-override]
        self,
        version: str,
        strategy: PEP440Strategy = "micro",
        *,
        zerover: bool = False,
        trim: bool = False,
    ) -> str:
        """Bump a PEP 440 version.

        Parameters:
            version: The version to bump.
            strategy: The part of the version to bump.
            zerover: Keep major version at zero, even for breaking changes.
            trim: Whether to trim all zeroes on the right after bumping.

        Returns:
            The bumped version.
        """
        pep440_version, prefix = parse_pep440(version)

        # Split into main part and pre/dev markers
        # (+alpha, +beta, +candidate, +dev).
        main_part, *predev = strategy.split("+")

        # Bump main part.
        if main_part == "epoch":
            pep440_version = pep440_version.bump_epoch()
        elif main_part == "release":
            pep440_version = pep440_version.bump_release(trim=trim)
        elif main_part == "major":
            # If major version is 0 and zerover is active, only bump minor.
            if pep440_version.major == 0 and zerover:
                pep440_version = pep440_version.bump_minor(trim=trim)
            else:
                pep440_version = pep440_version.bump_major(trim=trim)
        elif main_part == "minor":
            pep440_version = pep440_version.bump_minor(trim=trim)
        elif main_part in ("micro", "patch"):
            pep440_version = pep440_version.bump_micro(trim=trim)
        elif main_part == "pre":
            pep440_version = pep440_version.bump_pre()
        elif main_part == "alpha":
            pep440_version = pep440_version.bump_alpha()
        elif main_part == "beta":
            pep440_version = pep440_version.bump_beta()
        elif main_part == "candidate":
            pep440_version = pep440_version.bump_candidate()
        elif main_part == "post":
            pep440_version = pep440_version.bump_post()
        elif main_part == "dev":
            pep440_version = pep440_version.bump_dev()
        else:
            raise ValueError(f"Invalid strategy {main_part}, use one of {', '.join(self.strategies)}")

        # Dent to pre-release (+alpha, +beta, +candidate).
        if "alpha" in predev:
            pep440_version = pep440_version.dent_alpha()
        elif "beta" in predev:
            pep440_version = pep440_version.dent_beta()
        elif "candidate" in predev:
            pep440_version = pep440_version.dent_candidate()

        # Dent to dev-release (+dev).
        if "dev" in predev:
            pep440_version = pep440_version.dent_dev()

        # Return new version with preserved prefix.
        return prefix + str(pep440_version)


class SemVerBumper(VersionBumper):
    """SemVer version bumper."""

    initial: str = "0.0.0"
    """The initial version."""

    def __call__(  # ty:ignore[invalid-method-override]
        self,
        version: str,
        strategy: SemVerStrategy = "patch",
        *,
        zerover: bool = True,
    ) -> str:
        """Bump a SemVer version.

        Parameters:
            version: The version to bump.
            strategy: The bumping strategy to use.
            zerover: Keep major version at zero, even for breaking changes.

        Returns:
            The bumped version.
        """
        semver_version, prefix = parse_semver(version)
        if strategy == "major":
            if semver_version.major == 0 and zerover:
                semver_version = semver_version.bump_minor()
            else:
                semver_version = semver_version.bump_major()
        elif strategy == "minor":
            semver_version = semver_version.bump_minor()
        elif strategy == "patch":
            semver_version = semver_version.bump_patch()
        elif strategy == "release":
            semver_version = semver_version.bump_release()
        else:
            raise ValueError(f"Invalid strategy {strategy}, use one of {', '.join(self.strategies)}")
        return prefix + str(semver_version)


bump_pep440 = PEP440Bumper(PEP440Strategy.__args__)
"""Bump a PEP 440 version."""
bump_semver = SemVerBumper(SemVerStrategy.__args__)
"""Bump a SemVer version."""



"""Helpers for writing tests."""

from __future__ import annotations

import os
import random
import shutil
import subprocess
import uuid
from contextlib import contextmanager
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path


class GitRepo:
    """Test utility class to initalize and work with a Git repository."""

    _git_exec = shutil.which("git") or "git"

    def __init__(self, repo: Path) -> None:
        """Initialization the Git repository wrapper.

        Initializes a new Git repository under the given path.

        Parameters:
            repo: Path to the git repository.
        """
        self.path = repo
        self.path.mkdir(parents=True, exist_ok=True)
        self.git("init", "-b", "main")
        self.git("remote", "add", "origin", "git@github.com:example/example")
        self.first_hash = self.commit("chore: Initial repository creation")

    def git(self, *args: str) -> str:
        """Run a Git command in the repository.

        Parameters:
            *args: Arguments passed to the Git command.

        Returns:
            The output of the command.
        """
        return subprocess.check_output(  # noqa: S603
            [self._git_exec, "-C", str(self.path), *args],
            text=True,
        )

    def commit(self, message: str) -> str:
        """Create, add and commit a new file into the Git repository.

        Parameters:
            message: The commit message.

        Returns:
            The Git commit hash.
        """
        with self.path.joinpath(str(uuid.uuid4())).open("w", encoding="utf8") as fh:
            fh.write(str(random.randint(0, 1)))  # noqa: S311
        self.git("add", "-A")
        self.git("commit", "-m", message)
        return self.git("rev-parse", "HEAD").rstrip()

    def tag(self, tagname: str) -> None:
        """Create a new tag in the GIt repository.

        Parameters:
            tagname: The name of the new tag.
        """
        self.git("tag", tagname)

    def branch(self, branchname: str) -> None:
        """Create a new branch in the Git repository.

        Parameters:
            branchname: The name of the new branch.
        """
        self.git("branch", branchname)

    def checkout(self, branchname: str) -> None:
        """Checkout a branch.

        Parameters:
            branchname: The name of the branch.
        """
        self.git("checkout", branchname)

    def merge(self, branchname: str) -> str:
        """Merge a branch into the current branch, creating a new merge commit.

        Parameters:
            branchname: The name of the branch to merge.

        Returns:
            The Git commit hash of the merge commit.
        """
        self.git("merge", "--no-ff", "--commit", "-m", f"merge: Merge branch '{branchname}'", branchname)
        return self.git("rev-parse", "HEAD").rstrip()

    @contextmanager
    def enter(self) -> Iterator[None]:
        """Context manager to enter the repository directory."""
        old_cwd = os.getcwd()  # noqa: PTH109
        os.chdir(self.path)
        try:
            yield
        finally:
            os.chdir(old_cwd)



"""Tests for the `commit` module."""

from __future__ import annotations

import datetime

import pytest

from git_changelog import Commit


@pytest.mark.parametrize(
    ("body", "expected_trailers"),
    [
        ("t1: v1\nt2: v2", [("t1", "v1"), ("t2", "v2")]),  # ok
        ("body\n\nt1: v1\nt2: v2", [("t1", "v1"), ("t2", "v2")]),  # ok
        ("t1: v1\nt2:v2", []),  # missing space after colon
        ("t1: v1\nt2: v2\n\nf", []),  # trailers not last
        ("t1: v1\nt2 v2", []),  # not all trailers
        (
            "something: else\n\nt1: v1\nt2: v2",
            [("t1", "v1"), ("t2", "v2")],
        ),  # parse footer only
        ("t1: v1\nt1: v2", [("t1", "v1"), ("t1", "v2")]),  # multiple identical trailers
    ],
)
def test_parsing_trailers(body: str, expected_trailers: list[tuple[str, str]]) -> None:
    """Assert trailers are parsed correctly.

    Parameters:
        body: A commit message body.
        expected_trailers: The trailers we expect to be parsed.
    """
    commit = Commit(
        commit_hash="aaaaaaaa",
        subject="Summary",
        body=body.split("\n"),
        parse_trailers=True,
    )
    assert commit.trailers == expected_trailers


# YORE: Bump 3: Remove block.
def test_trailers_emit_deprecation_warnings() -> None:
    """Trailers used as a dictionary emit deprecation warnings."""
    commit = Commit(
        commit_hash="aaaaaaaa",
        subject="Summary",
        parse_trailers=True,
    )
    with pytest.warns(DeprecationWarning, match="Trailers are now a list of 2-tuples."):
        assert not commit.trailers.keys()  # ty:ignore[unresolved-attribute]
    with pytest.warns(DeprecationWarning, match="Trailers are now a list of 2-tuples."):
        assert not commit.trailers.values()  # ty:ignore[unresolved-attribute]
    with pytest.warns(DeprecationWarning, match="Trailers are now a list of 2-tuples."):
        assert not commit.trailers.items()  # ty:ignore[unresolved-attribute]
    with pytest.warns(DeprecationWarning, match="Trailers are now a list of 2-tuples."):
        assert not commit.trailers.get("key")  # ty:ignore[unresolved-attribute]
    with (
        pytest.warns(
            DeprecationWarning,
            match="Getting a trailer with a string key will stop being supported in version 3. Instead, use an integer index, or iterate.",
        ),
        pytest.raises(KeyError),
    ):
        assert not commit.trailers["key"]  # ty:ignore[invalid-argument-type]


def test_tzinfo() -> None:
    """Keep timezone info from commit."""
    commit = Commit(
        commit_hash="aaaaaaaa",
        subject="Summary",
        parse_trailers=True,
        author_date="1772727280 -0500",  # 2026-03-05 11:13:27 -0500
        committer_date="1772830826 -0500",  # 2026-03-06 16:00:26 -0500
    )
    tz = datetime.timezone(datetime.timedelta(days=-1, seconds=68400))
    assert commit.author_date == datetime.datetime(2026, 3, 5, 11, 14, 40, tzinfo=tz)
    assert commit.committer_date == datetime.datetime(2026, 3, 6, 16, 0, 26, tzinfo=tz)
